"""Module with functions that encode classical data into quantum circuits."""

import math
from inspect import signature
from re import finditer
from typing import List, Optional, Union

import numpy as np
from scipy.special import binom

from qibo import gates
from qibo.backends import _check_backend
from qibo.config import raise_error
from qibo.models.circuit import Circuit


def comp_basis_encoder(
    basis_element: Union[int, str, list, tuple], nqubits: Optional[int] = None, **kwargs
):
    """Create circuit that performs encoding of bitstrings into computational basis states.

    Args:
        basis_element (int or str or list or tuple): bitstring to be encoded.
            If ``int``, ``nqubits`` must be specified.
            If ``str``, must be composed of only :math:`0`s and :math:`1`s.
            If ``list`` or ``tuple``, must be composed of :math:`0`s and
            :math:`1`s as ``int`` or ``str``.
        nqubits (int, optional): total number of qubits in the circuit.
            If ``basis_element`` is ``int``, ``nqubits`` must be specified.
            If ``nqubits`` is ``None``, ``nqubits`` defaults to length of ``basis_element``.
            Defaults to ``None``.
        kwargs (dict, optional): Additional arguments used to initialize a Circuit object.
            For details, see the documentation of :class:`qibo.models.circuit.Circuit`.

    Returns:
       :class:`qibo.models.circuit.Circuit`: Circuit encoding computational basis element.
    """
    if not isinstance(basis_element, (int, str, list, tuple)):
        raise_error(
            TypeError,
            "basis_element must be either type int or str or list or tuple, "
            + f"but it is type {type(basis_element)}.",
        )

    if isinstance(basis_element, (str, list, tuple)):
        if any(elem not in ["0", "1", 0, 1] for elem in basis_element):
            raise_error(ValueError, "all elements must be 0 or 1.")

    if nqubits is not None and not isinstance(nqubits, int):
        raise_error(
            TypeError, f"``nqubits`` must be type int, but it is type {type(nqubits)}."
        )

    if nqubits is None:
        if isinstance(basis_element, int):
            raise_error(
                ValueError,
                "``nqubits`` must be specified when ``basis_element`` is type int.",
            )
        else:
            nqubits = len(basis_element)

    if isinstance(basis_element, int):
        basis_element = f"{basis_element:0{nqubits}b}"

    if isinstance(basis_element, (str, tuple)):
        basis_element = list(basis_element)

    basis_element = list(map(int, basis_element))

    circuit = Circuit(nqubits, **kwargs)
    for qubit, elem in enumerate(basis_element):
        if elem == 1:
            circuit.add(gates.X(qubit))

    return circuit


def phase_encoder(data, rotation: str = "RY", backend=None, **kwargs):
    """Create circuit that performs the phase encoding of ``data``.

    Args:
        data (ndarray or list): :math:`1`-dimensional array of phases to be loaded.
        rotation (str, optional): If ``"RX"``, uses :class:`qibo.gates.gates.RX` as rotation.
            If ``"RY"``, uses :class:`qibo.gates.gates.RY` as rotation.
            If ``"RZ"``, uses :class:`qibo.gates.gates.RZ` as rotation.
            Defaults to ``"RY"``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend. Defaults to ``None``.
        kwargs (dict, optional): Additional arguments used to initialize a Circuit object.
            For details, see the documentation of :class:`qibo.models.circuit.Circuit`.

    Returns:
        :class:`qibo.models.circuit.Circuit`: Circuit that loads ``data`` in phase encoding.
    """
    if not isinstance(rotation, str):
        raise_error(
            TypeError,
            f"``rotation`` must be type str, but it is type {type(rotation)}.",
        )

    backend = _check_backend(backend)

    if isinstance(data, list):
        # TODO: Fix this mess with qibo native dtypes
        try:
            type_test = data[0].dtype
        except AttributeError:  # pragma: no cover
            type_test = type(data[0])

        data = backend.cast(data, dtype=type_test)

    if rotation not in ["RX", "RY", "RZ"]:
        raise_error(ValueError, f"``rotation`` {rotation} not found.")

    nqubits = len(data)
    gate = getattr(gates, rotation.upper())

    circuit = Circuit(nqubits, **kwargs)
    circuit.add(gate(qubit, 0.0) for qubit in range(nqubits))
    circuit.set_parameters(data)

    return circuit


def sparse_encoder(
    data, method: str = "li", nqubits: int = None, backend=None, **kwargs
):
    """Create circuit that encodes :math:`1`-dimensional data in a subset of amplitudes
    of the computational basis.

    Consider a sparse-access model, where for a data vector
    :math:`\\mathbf{x} \\in \\mathbb{C}^{d}`, with :math:`d = 2^{n}` and
    :math:`s` non-zero amplitudes, one has access to the data vector
    :math:`\\mathbf{y}` of the form

    .. math::
        \\mathbf{y} = \\left\\{ (b_{1}, x_{1}), \\, \\dots, \\, (b_{s}, x_{s}) \\right\\} \\, ,


    where :math:`\\{x_{j}\\}_{j\\in[s]}` is the non-zero components of :math:`\\mathbf{x}`
    and :math:`\\{b_{j}\\}_{j\\in[s]}` is the set of addresses associated with these values.
    Then, this function generates a quantum circuit  :math:`s\\text{-}\\mathrm{Load}` that encodes
    :math:`\\mathbf{x}` in the amplitudes of an :math:`n`-qubit quantum state as

    .. math::
        s\\text{-}\\mathrm{Load}(\\mathbf{y}) \\, \\ket{0}^{\\otimes \\, n} = \\sum_{j\\in[s]} \\,
            \\frac{x_{j}}{\\|\\mathbf{x}\\|_{2}} \\, \\ket{b_{j}} \\, ,

    where :math:`\\|\\cdot\\|_{2}` is the Euclidean norm.

    The resulting circuit parametrizes ``data`` in hyperspherical coordinates
    in the :math:`(2^{n} - 1)`-unit sphere.


    Args:
        data (ndarray or list or zip): sequence of tuples of the form :math:`(b_{j}, x_{j})`.
            The addresses :math:`b_{j}` can be either integers or in bitstring
            format of size :math:`n`.
        method (str, optional): method to be used, either ``li`` or ``farias``. They refer to
            methods in references [1] and [2], respectively.
            Defaults to ``li``
        nqubits (int, optional): total number of qubits in the system.
            To be used when :math:`b_j` are integers. If :math:`b_j` are strings and
            ``nqubits`` is ``None``, defaults to the length of the strings :math:`b_{j}`.
            Defaults to ``None``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend. Defaults to ``None``.
        kwargs (dict, optional): Additional arguments used to initialize a Circuit object.
            For details, see the documentation of :class:`qibo.models.circuit.Circuit`.

    Returns:
        :class:`qibo.models.circuit.Circuit`: Circuit that loads sparse :math:`\\mathbf{x}`.

    References:
        1. L. Li, and J. Luo,
        *Nearly Optimal Circuit Size for Sparse Quantum State Preparation*
        `arXiv:2406.16142 (2024) <https://doi.org/10.48550/arXiv.2406.16142>`_.

        2. R. M. S. Farias, T. O. Maciel, G. Camilo, R. Lin, S. Ramos-Calderer, and L. Aolita,
        *Quantum encoder for fixed-Hamming-weight subspaces*,
        `Phys. Rev. Applied 23, 044014 (2025) <https://doi.org/10.1103/PhysRevApplied.23.044014>`_.

        3. `Hyperpherical coordinates <https://en.wikipedia.org/wiki/N-sphere>`_.
    """
    backend = _check_backend(backend)

    if isinstance(data, zip):
        data = list(data)

    # TODO: Fix this mess with qibo native dtypes
    try:
        type_test = bool("int" in str(data[0][0].dtype))
    except AttributeError:
        type_test = bool("int" in str(type(data[0][0])))

    if type_test and nqubits is None:
        raise_error(
            ValueError,
            "``nqubits`` must be specified when computational basis states are "
            + "indidated by integers.",
        )

    if isinstance(data[0][0], str) and nqubits is None:
        nqubits = len(data[0][0])

    if method not in ("li", "farias"):
        raise_error(
            ValueError,
            f"``method`` should be either ``li`` or ``farias``, but it is {method}",
        )

    func = _sparse_encoder_farias if method == "farias" else _sparse_encoder_li

    return func(data, nqubits, backend, **kwargs)


def _sparse_encoder_li(data, nqubits: int, backend=None, **kwargs):
    """Create circuit that encodes :math:`1`-dimensional data in a subset of amplitudes
    of the computational basis.

    Consider a sparse-access model, where for a data vector
    :math:`\\mathbf{x} \\in \\mathbb{C}^{d}`, with :math:`d = 2^{n}` and
    :math:`s` non-zero amplitudes, one has access to the data vector
    :math:`\\mathbf{y}` of the form

    .. math::
        \\mathbf{y} = \\left\\{ (b_{1}, x_{1}), \\, \\dots, \\, (b_{s}, x_{s}) \\right\\} \\, ,


    where :math:`\\{x_{j}\\}_{j\\in[s]}` are the non-zero components of :math:`\\mathbf{x}`
    and :math:`\\{b_{j}\\}_{j\\in[s]}` is the set of addresses associated with these values.
    Then, this function generates a quantum circuit  :math:`s\\text{-}\\mathrm{Load}` that encodes
    :math:`\\mathbf{x}` in the amplitudes of an :math:`n`-qubit quantum state as

    .. math::
        s\\text{-}\\mathrm{Load}(\\mathbf{y}) \\, \\ket{0}^{\\otimes \\, n} = \\sum_{j\\in[s]} \\,
            \\frac{x_{j}}{\\|\\mathbf{x}\\|_{2}} \\, \\ket{b_{j}} \\, ,

    where :math:`\\|\\cdot\\|_{2}` is the l2-norm.

    The resulting circuit parametrizes ``data`` in hyperspherical coordinates
    in the :math:`(2^{n} - 1)`-unit sphere.


    Args:
        data (ndarray or list or zip): sequence of tuples of the form :math:`(b_{j}, x_{j})`.
            The addresses :math:`b_{j}` can be either integers or in bitstring
            format of size :math:`n`.
        nqubits (int, optional): total number of qubits in the system.
            To be used when :math:`b_j` are integers. If :math:`b_j` are strings and
            ``nqubits`` is ``None``, defaults to the length of the strings :math:`b_{j}`.
            Defaults to ``None``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend. Defaults to ``None``.
        kwargs (dict, optional): Additional arguments used to initialize a Circuit object.
            For details, see the documentation of :class:`qibo.models.circuit.Circuit`.

    Returns:
        :class:`qibo.models.circuit.Circuit`: Circuit that loads sparse :math:`\\mathbf{x}`.

    References:
        1. L. Li, and J. Luo,
        *Nearly Optimal Circuit Size for Sparse Quantum State Preparation*
        `arXiv:2406.16142 (2024) <https://doi.org/10.48550/arXiv.2406.16142>`_.

        2. `Hyperpherical coordinates <https://en.wikipedia.org/wiki/N-sphere>`_.
    """
    backend = _check_backend(backend)
    data_sorted, bitstrings_sorted = _sort_data_sparse(data, nqubits, backend)
    bitstrings_sorted = backend.cast(
        [int("".join(map(str, string)), 2) for string in bitstrings_sorted],
        dtype=backend.np.int8,
    )

    dim = len(data_sorted)
    sigma = np.arange(2**nqubits)

    flag = backend.np.zeros(dim, dtype=backend.np.int8)
    indexes = list(
        backend.to_numpy(bitstrings_sorted[bitstrings_sorted < dim]).astype(int)
    )
    flag[indexes] = 1

    data_binary = []
    for bi_int, xi in zip(bitstrings_sorted, data_sorted):
        bi_int = int(bi_int)
        if bi_int >= dim:
            for k in range(dim):
                if flag[k] == 0:
                    flag[k] = 1
                    sigma[[bi_int, k]] = [k, bi_int]
                    data_binary.append((k, xi))
                    break
        else:
            data_binary.append((bi_int, xi))

    sigma = list(sigma)

    # binary enconder on \sum_i = xi |sigma^{-1}(b_i)>
    circuit_binary = sparse_encoder(
        data_binary, method="farias", nqubits=nqubits, backend=backend, **kwargs
    )
    circuit_permutation = permutation_synthesis(sigma, **kwargs)

    return circuit_binary + circuit_permutation


def _sparse_encoder_farias(data, nqubits: int, backend=None, **kwargs):
    """Create circuit that encodes :math:`1`-dimensional data in a subset of amplitudes
    of the computational basis.

    Consider a sparse-access model, where for a data vector
    :math:`\\mathbf{x} \\in \\mathbb{C}^{d}`, with :math:`d = 2^{n}` and
    :math:`s` non-zero amplitudes, one has access to the data vector
    :math:`\\mathbf{y}` of the form

    .. math::
        \\mathbf{y} = \\left\\{ (b_{1}, x_{1}), \\, \\dots, \\, (b_{s}, x_{s}) \\right\\} \\, ,


    where :math:`\\{x_{j}\\}_{j\\in[s]}` are the non-zero components of :math:`\\mathbf{x}`
    and :math:`\\{b_{j}\\}_{j\\in[s]}` is the set of addresses associated with these values.
    Then, this function generates a quantum circuit  :math:`s\\text{-}\\mathrm{Load}` that encodes
    :math:`\\mathbf{x}` in the amplitudes of an :math:`n`-qubit quantum state as

    .. math::
        s\\text{-}\\mathrm{Load}(\\mathbf{y}) \\, \\ket{0}^{\\otimes \\, n} = \\sum_{j\\in[s]} \\,
            \\frac{x_{j}}{\\|\\mathbf{x}\\|_{2}} \\, \\ket{b_{j}} \\, ,

    where :math:`\\|\\cdot\\|_{2}` is the l2-norm.

    Resulting circuit parametrizes ``data`` in hyperspherical coordinates
    in the :math:`(2^{n} - 1)`-unit sphere.


    Args:
        data (ndarray or list or zip): sequence of tuples of the form :math:`(b_{j}, x_{j})`.
            The addresses :math:`b_{j}` can be either integers or in bitstring
            format of size :math:`n`.
        nqubits (int, optional): total number of qubits in the system.
            To be used when :math:`b_j` are integers. If :math:`b_j` are strings and
            ``nqubits`` is ``None``, defaults to the length of the strings :math:`b_{j}`.
            Defaults to ``None``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend. Defaults to ``None``.
        kwargs (dict, optional): Additional arguments used to initialize a Circuit object.
            For details, see the documentation of :class:`qibo.models.circuit.Circuit`.

    Returns:
        :class:`qibo.models.circuit.Circuit`: Circuit that loads sparse :math:`\\mathbf{x}`.

    References:
        1. R. M. S. Farias, T. O. Maciel, G. Camilo, R. Lin, S. Ramos-Calderer, and L. Aolita,
        *Quantum encoder for fixed-Hamming-weight subspaces*,
        `Phys. Rev. Applied 23, 044014 (2025) <https://doi.org/10.1103/PhysRevApplied.23.044014>`_.

        2. `Hyperpherical coordinates <https://en.wikipedia.org/wiki/N-sphere>`_.
    """
    from qibo.quantum_info.utils import (  # pylint: disable=import-outside-toplevel
        hamming_distance,
        hamming_weight,
    )

    backend = _check_backend(backend)

    if isinstance(data, zip):
        data = list(data)

    # TODO: Fix this mess with qibo native dtypes
    try:
        type_test = bool("int" in str(data[0][0].dtype))
    except AttributeError:
        type_test = bool("int" in str(type(data[0][0])))

    if type_test and nqubits is None:
        raise_error(
            ValueError,
            "``nqubits`` must be specified when computational basis states are "
            + "indidated by integers.",
        )

    if isinstance(data[0][0], str) and nqubits is None:
        nqubits = len(data[0][0])

    _data_test = data[0][1]
    _data_test = (
        _data_test.dtype if "array" in str(type(_data_test)) else type(_data_test)
    )

    complex_data = bool("complex" in str(_data_test))

    # sort data by HW of the bitstrings
    data_sorted, bitstrings_sorted = _sort_data_sparse(data, nqubits, backend)
    # calculate phases
    _data_sorted = backend.np.abs(data_sorted) if complex_data else data_sorted
    thetas = _generate_rbs_angles(
        _data_sorted, architecture="diagonal", backend=backend
    )
    phis = backend.np.zeros(len(thetas) + 1)
    if complex_data:
        phis[0] = _angle_mod_two_pi(-backend.np.angle(data_sorted[0]))
        for k in range(1, len(phis)):
            phis[k] = _angle_mod_two_pi(
                -backend.np.angle(data_sorted[k]) + backend.np.sum(phis[:k])
            )
    phis = backend.cast(phis, dtype=phis[0].dtype)

    # marking qubits that have suffered the action of a gate
    initial_string = [int(bit) for bit in bitstrings_sorted[0]]
    circuit = comp_basis_encoder(initial_string, nqubits=nqubits, **kwargs)
    touched_qubits = list(np.nonzero(initial_string)[0])

    for b_1, b_0, theta, phi in zip(
        bitstrings_sorted[1:], bitstrings_sorted[:-1], thetas, phis
    ):
        hw_0, hw_1 = hamming_weight(b_0), hamming_weight(b_1)
        distance = hamming_distance(b_1, b_0)
        difference = b_1 - b_0

        ones, new_ones = (
            list(backend.np.argsort(b_0)[-hw_0:]),
            list(backend.np.argsort(b_1)[-hw_1:]),
        )
        ones, new_ones = {int(elem) for elem in ones}, {int(elem) for elem in new_ones}
        controls = (set(ones) & set(new_ones)) & set(touched_qubits)

        gate = _get_gate_sparse(
            distance,
            difference,
            touched_qubits,
            complex_data,
            controls,
            hw_0,
            hw_1,
            theta,
            phi,
        )
        circuit.add(gate)

    if complex_data:
        hw_0 = hamming_weight(bitstrings_sorted[-2])
        hw_1 = hamming_weight(bitstrings_sorted[-1])
        correction = _get_phase_gate_correction_sparse(
            bitstrings_sorted[-1],
            bitstrings_sorted[-2],
            nqubits,
            data_sorted[-1],
            data_sorted[-2],
            circuit,
            phis,
        )
        if hw_1 == nqubits and hw_0 == nqubits - 1:
            circuit.queue = correction
        else:
            circuit.add(correction)

    return circuit


def binary_encoder(
    data, parametrization: str = "hyperspherical", backend=None, **kwargs
):
    """Create circuit that encodes :math:`1`-dimensional data in all amplitudes
    of the computational basis.

    Given data vector :math:`\\mathbf{x} \\in \\mathbb{C}^{d}`, with :math:`d = 2^{n}`,
    this function generates a quantum circuit :math:`\\mathrm{Load}` that encodes
    :math:`\\mathbf{x}` in the amplitudes of an :math:`n`-qubit quantum state as

    .. math::
        \\mathrm{Load}(\\mathbf{x}) \\, \\ket{0}^{\\otimes \\, n} = \\sum_{j=0}^{d-1} \\,
            \\frac{x_{j}}{\\|\\mathbf{x}\\|_{F}} \\, \\ket{b_{j}} \\, ,

    where :math:`b_{j} \\in \\{0, \\, 1\\}^{\\otimes \\, n}` is the :math:`n`-bit representation
    of the integer :math:`j`, :math:`\\|\\cdot\\|_{F}` is the Frobenius norm.

    Resulting circuit parametrizes ``data`` in either ``hyperspherical`` or ``Hopf`` coordinates
    in the :math:`(2^{n} - 1)`-unit sphere.

    Args:
        data (ndarray): :math:`1`-dimensional array or length :math:`d = 2^{n}`
            to be loaded in the amplitudes of a :math:`n`-qubit quantum state.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend. Defaults to ``None``.
        kwargs (dict, optional): Additional arguments used to initialize a Circuit object.
            For details, see the documentation of :class:`qibo.models.circuit.Circuit`.

    Returns:
        :class:`qibo.models.circuit.Circuit`: Circuit that loads ``data`` in binary encoding.

    References:
        1. R. M. S. Farias, T. O. Maciel, G. Camilo, R. Lin, S. Ramos-Calderer, and L. Aolita,
        *Quantum encoder for fixed-Hamming-weight subspaces*
        `Phys. Rev. Applied 23, 044014 (2025) <https://doi.org/10.1103/PhysRevApplied.23.044014>`_.

        2. `Hyperpherical coordinates <https://en.wikipedia.org/wiki/N-sphere>`_.

        3. H. S. Cohl, *Fourier, Gegenbauer and Jacobi expansions for a power-law fundamental
        solution of the polyharmonic equation and polyspherical addition theorems*, `Symmetry,
        Integrability and Geometry: Methods and Applications 10.3842/sigma.2013.042 (2013)
        <https://arxiv.org/abs/1209.6047>`_.
    """
    backend = _check_backend(backend)

    dims = len(data)
    nqubits = float(np.log2(dims))
    if not nqubits.is_integer():
        raise_error(ValueError, "`data` size must be a power of 2.")
    nqubits = int(nqubits)

    complex_data = bool(
        "complex" in str(data.dtype)
    )  # backend-agnostic way of checking the dtype

    if parametrization == "hopf":
        return _binary_encoder_hopf(
            data, nqubits, complex_data=complex_data, backend=backend, **kwargs
        )

    return _binary_encoder_hyperspherical(
        data, nqubits, complex_data=complex_data, backend=backend, **kwargs
    )


def unary_encoder(data, architecture: str = "tree", backend=None, **kwargs):
    """Create circuit that performs the (deterministic) unary encoding of ``data``.

    Args:
        data (ndarray): :math:`1`-dimensional array of data to be loaded.
        architecture(str, optional): circuit architecture used for the unary loader.
            If ``diagonal``, uses a ladder-like structure.
            If ``tree``, uses a binary-tree-based structure.
            Defaults to ``tree``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend. Defaults to ``None``.
        kwargs (dict, optional): Additional arguments used to initialize a Circuit object.
            For details, see the documentation of :class:`qibo.models.circuit.Circuit`.

    Returns:
        :class:`qibo.models.circuit.Circuit`: Circuit that loads ``data`` in unary representation.
    """
    backend = _check_backend(backend)

    if isinstance(data, list):
        data = backend.cast(data, dtype=type(data[0]))

    if not isinstance(architecture, str):
        raise_error(
            TypeError,
            f"``architecture`` must be type str, but it is type {type(architecture)}.",
        )

    if architecture not in ["diagonal", "tree"]:
        raise_error(ValueError, f"``architecture`` {architecture} not found.")

    if architecture == "tree" and not math.log2(data.shape[0]).is_integer():
        raise_error(
            ValueError,
            "When ``architecture = 'tree'``, len(data) must be a power of 2. "
            + f"However, it is {len(data)}.",
        )

    nqubits = len(data)

    circuit = Circuit(nqubits, **kwargs)
    circuit.add(gates.X(nqubits - 1))
    circuit_rbs, _ = _generate_rbs_pairs(nqubits, architecture=architecture, **kwargs)
    circuit += circuit_rbs

    # calculating phases and setting circuit parameters
    phases = _generate_rbs_angles(data, architecture, nqubits, backend=backend)
    circuit.set_parameters(phases)

    return circuit


def unary_encoder_random_gaussian(
    nqubits: int, architecture: str = "tree", seed=None, backend=None, **kwargs
):
    """Create a circuit that performs the unary encoding of a random Gaussian state.

    At depth :math:`h` of the tree architecture, the angles :math:`\\theta_{k}
    \\in [0, 2\\pi]` of the the gates :math:`RBS(\\theta_{k})` are sampled from
    the following probability density function:

    .. math::
        p_{h}(\\theta) = \\frac{1}{2} \\, \\frac{\\Gamma(2^{h-1})}{\\Gamma^{2}(2^{h-2})} \\,
            \\left|\\sin(\\theta) \\, \\cos(\\theta)\\right|^{2^{h-1} - 1} \\, ,

    where :math:`\\Gamma(\\cdot)` is the
    `Gamma function <https://en.wikipedia.org/wiki/Gamma_function>`_.

    Args:
        nqubits (int): number of qubits.
        architecture(str, optional): circuit architecture used for the unary loader.
            If ``tree``, uses a binary-tree-based structure.
            Defaults to ``tree``.
        seed (int or :class:`numpy.random.Generator`, optional): Either a generator of
            random numbers or a fixed seed to initialize a generator. If ``None``,
            initializes a generator with a random seed. Defaults to ``None``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend. Defaults to ``None``.
        kwargs (dict, optional): Additional arguments used to initialize a Circuit object.
            For details, see the documentation of :class:`qibo.models.circuit.Circuit`.

    Returns:
        :class:`qibo.models.circuit.Circuit`: Circuit that loads a random Gaussian
        array in unary representation.

    References:
        1. A. Bouland, A. Dandapani, and A. Prakash, *A quantum spectral method for simulating
        stochastic processes, with applications to Monte Carlo*.
        `arXiv:2303.06719v1 [quant-ph] <https://arxiv.org/abs/2303.06719>`_
    """
    backend = _check_backend(backend)

    if not isinstance(nqubits, int):
        raise_error(
            TypeError, f"nqubits must be type int, but it is type {type(nqubits)}."
        )

    if nqubits <= 0.0:
        raise_error(
            ValueError, f"nqubits must be a positive integer, but it is {nqubits}."
        )

    if not isinstance(architecture, str):
        raise_error(
            TypeError,
            f"``architecture`` must be type str, but it is type {type(architecture)}.",
        )

    if architecture != "tree":
        raise_error(
            NotImplementedError,
            "Currently, this function only accepts ``architecture=='tree'``.",
        )

    if not math.log2(nqubits).is_integer():
        raise_error(ValueError, f"nqubits must be a power of 2, but it is {nqubits}.")

    if (
        seed is not None
        and not isinstance(seed, int)
        and not isinstance(seed, np.random.Generator)
    ):
        raise_error(
            TypeError, "seed must be either type int or numpy.random.Generator."
        )

    from qibo.quantum_info.random_ensembles import (  # pylint: disable=C0415
        _ProbabilityDistributionGaussianLoader,
    )

    local_state = (
        np.random.default_rng(seed) if seed is None or isinstance(seed, int) else seed
    )

    sampler = _ProbabilityDistributionGaussianLoader(
        a=0, b=2 * math.pi, seed=local_state
    )

    circuit = Circuit(nqubits, **kwargs)
    circuit.add(gates.X(nqubits - 1))
    circuit_rbs, pairs_rbs = _generate_rbs_pairs(nqubits, architecture, **kwargs)
    circuit += circuit_rbs

    phases = []
    for depth, row in enumerate(pairs_rbs, 1):
        phases.extend(sampler.rvs(depth=depth, size=len(row)))

    phases = backend.cast(phases, dtype=type(phases[0]))

    circuit.set_parameters(phases)

    return circuit


def hamming_weight_encoder(
    data,
    nqubits: int,
    weight: int,
    full_hwp: bool = False,
    optimize_controls: bool = True,
    phase_correction: bool = True,
    initial_string=None,
    backend=None,
    **kwargs,
):
    """Create circuit that encodes ``data`` in the Hamming-weight-:math:`k` basis of ``nqubits``.

    Let :math:`\\mathbf{x}` be a :math:`1`-dimensional array of size :math:`d = \\binom{n}{k}` and
    :math:`B_{k} \\equiv \\{ \\ket{b_{j}} : b_{j} \\in \\{0, 1\\}^{\\otimes n} \\,\\, \\text{and}
    \\,\\, |b_{j}| = k \\}` be a set of :math:`d` computational basis states of :math:`n` qubits
    that are represented by bitstrings of Hamming weight :math:`k`. Then, an amplitude encoder
    in the basis :math:`B_{k}` is an :math:`n`-qubit parameterized quantum circuit
    :math:`\\operatorname{Load}_{B_{k}}` such that

    .. math::
        \\operatorname{Load}(\\mathbf{x}) \\, \\ket{0}^{\\otimes n} = \\frac{1}{\\|\\mathbf{x}\\|}
            \\, \\sum_{j = 1}^{d} \\, x_{j} \\, \\ket{b_{j}}

    Args:
        data (ndarray): :math:`1`-dimensional array of data to be loaded.
        nqubits (int): number of qubits.
        weight (int): Hamming weight that defines the subspace in which ``data`` will be encoded.
        full_hwp (bool, optional): if ``False``, includes Pauli-:math:`X` gates that prepare the
            first bitstring of Hamming weight ``k = weight``. If ``True``, circuit is full
            Hamming weight preserving. Defaults to ```False``.
        optimize_controls (bool, optional): if ``True``, removes unnecessary controlled operations.
            Defaults to ``True``.
        phase_correction (bool, optional): To be used when ``data`` is complex-valued.
            If ``True``, adds a controlled-$\\mathrm{RZ}$ gate to the end of the circuit,
            adding a final phase correction. If ``False``, gate is not added. Defaults to ``True``.
        initial_string (ndarray, optional): Array containing the desired initial bitstring of
            Hamming ``weight`` $k$. If ``None``, defaults to $\\ket{1^{k}0^{n-k}}$.
            Defaults to ``None``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend. Defaults to ``None``.
        kwargs (dict, optional): Additional arguments used to initialize a Circuit object.
            For details, see the documentation of :class:`qibo.models.circuit.Circuit`.

    Returns:
        :class:`qibo.models.circuit.Circuit`: Circuit that loads ``data`` in
        Hamming-weight-:math:`k` representation.

    References:
        1. R. M. S. Farias, T. O. Maciel, G. Camilo, R. Lin, S. Ramos-Calderer, and L. Aolita,
        *Quantum encoder for fixed-Hamming-weight subspaces*
        `Phys. Rev. Applied 23, 044014 (2025) <https://doi.org/10.1103/PhysRevApplied.23.044014>`_.
    """
    backend = _check_backend(backend)

    complex_data = bool("complex" in str(data.dtype))

    if initial_string is None:
        initial_string = np.array([1] * weight + [0] * (nqubits - weight))
    bitstrings, targets_and_controls = _ehrlich_algorithm(initial_string)

    # sort data such that the encoding is performed in lexicographical order
    lex_order = [int(string, 2) for string in bitstrings]
    lex_order_sorted = np.sort(np.copy(lex_order))
    lex_order = [np.where(lex_order_sorted == num)[0][0] for num in lex_order]
    data = data[lex_order]
    del lex_order, lex_order_sorted

    # Calculate all gate phases necessary to encode the amplitudes.
    _data = backend.np.abs(data) if complex_data else data
    thetas = _generate_rbs_angles(_data, architecture="diagonal", backend=backend)
    phis = backend.np.zeros(len(thetas) + 1)
    if complex_data:
        phis[0] = _angle_mod_two_pi(-backend.np.angle(data[0]))
        for k in range(1, len(phis)):
            phis[k] = _angle_mod_two_pi(
                -backend.np.angle(data[k]) + backend.np.sum(phis[:k])
            )

    last_qubit = nqubits - 1

    circuit = Circuit(nqubits, **kwargs)
    if not full_hwp:
        circuit.add(
            gates.X(qubit) for qubit in range(last_qubit, last_qubit - weight, -1)
        )

    if optimize_controls:
        indices = [
            int(binom(nqubits - j, weight - j)) - 1 for j in range(weight - 1, 0, -1)
        ]

    for k, ((targets, controls), theta, phi) in enumerate(
        zip(targets_and_controls, thetas, phis)
    ):
        targets = list(last_qubit - np.asarray(targets))
        controls = list(last_qubit - np.asarray(controls))
        controls.sort()

        if optimize_controls:
            controls = list(np.asarray(controls)[k >= np.asarray(indices)])

        gate = _get_gate(
            [targets[0]],
            [targets[1]],
            controls,
            theta,
            phi,
            complex_data,
        )
        circuit.add(gate)

    if complex_data and phase_correction:
        circuit.add(_get_phase_gate_correction(bitstrings[-1], phis[-1]))

    return circuit


def entangling_layer(
    nqubits: int,
    architecture: str = "diagonal",
    entangling_gate: Union[str, gates.Gate] = "CNOT",
    closed_boundary: bool = False,
    **kwargs,
):
    """Create a layer of two-qubit entangling gates.

    If the chosen gate is a parametrized gate, all phases are set to :math:`0.0`.

    Args:
        nqubits (int): Total number of qubits in the circuit.
        architecture (str, optional): Architecture of the entangling layer.
            In alphabetical order, options are ``"diagonal"``, ``"even_layer"``,
            ``"next_nearest"``, ``"odd_layer"``, ``"pyramid"``, ``"shifted"``,
            ``"v"``, and ``"x"``. The ``"x"`` architecture is only defined for an even number
            of qubits. Defaults to ``"diagonal"``.
        entangling_gate (str or :class:`qibo.gates.Gate`, optional): Two-qubit gate to be used
            in the entangling layer. If ``entangling_gate`` is a parametrized gate,
            all phases are initialized as :math:`0.0`. Defaults to  ``"CNOT"``.
        closed_boundary (bool, optional): If ``True`` and ``architecture not in
            ["pyramid", "v", "x"]``, adds a closed-boundary condition to the entangling layer.
            Defaults to ``False``.
        kwargs (dict, optional): Additional arguments used to initialize a Circuit object.
            For details, see the documentation of :class:`qibo.models.circuit.Circuit`.

    Returns:
        :class:`qibo.models.circuit.Circuit`: Circuit containing layer of two-qubit gates.
    """

    if not isinstance(nqubits, int):
        raise_error(
            TypeError, f"nqubits must be type int, but it is type {type(nqubits)}."
        )

    if nqubits <= 0.0:
        raise_error(
            ValueError, f"nqubits must be a positive integer, but it is {nqubits}."
        )

    if not isinstance(architecture, str):
        raise_error(
            TypeError,
            f"``architecture`` must be type str, but it is type {type(architecture)}.",
        )

    if architecture not in [
        "diagonal",
        "even_layer",
        "next_nearest",
        "odd_layer",
        "pyramid",
        "shifted",
        "v",
        "x",
    ]:
        raise_error(
            NotImplementedError,
            f"``architecture`` {architecture} not found.",
        )

    if architecture == "x" and nqubits % 2 != 0.0:
        raise_error(
            ValueError, "``x`` architecture only defined for an even number of qubits."
        )

    if not isinstance(closed_boundary, bool):
        raise_error(
            TypeError,
            f"``closed_boundary`` must be type bool, but it is type {type(closed_boundary)}.",
        )

    gate = (
        getattr(gates, entangling_gate)
        if isinstance(entangling_gate, str)
        else entangling_gate
    )

    if gate.__name__ == "GeneralizedfSim":
        raise_error(
            NotImplementedError,
            "This function does not support the ``GeneralizedfSim`` gate.",
        )

    if architecture in ["next_nearest", "pyramid", "v", "x"]:
        circuit = _non_trivial_layers(
            nqubits,
            architecture=architecture,
            entangling_gate=entangling_gate,
            closed_boundary=closed_boundary,
            **kwargs,
        )
    else:
        # Finds the correct number of parameters to initialize the gate class.
        parameters = list(signature(gate).parameters)

        if "q2" in parameters:
            raise_error(
                NotImplementedError, "This function does not accept three-qubit gates."
            )

        # If gate is parametrized, sets all angles to 0.0
        parameters = (0.0,) * (len(parameters) - 3) if len(parameters) > 2 else None

        circuit = Circuit(nqubits, **kwargs)

        if architecture == "diagonal":
            qubits = range(nqubits - 1)
        elif architecture == "even_layer":
            qubits = range(0, nqubits - 1, 2)
        elif architecture == "odd_layer":
            qubits = range(1, nqubits - 1, 2)
        else:
            qubits = tuple(range(0, nqubits - 1, 2)) + tuple(range(1, nqubits - 1, 2))

        circuit.add(
            _parametrized_two_qubit_gate(gate, qubit, qubit + 1, parameters)
            for qubit in qubits
        )

        if closed_boundary:
            circuit.add(_parametrized_two_qubit_gate(gate, nqubits - 1, 0, parameters))

    return circuit


def ghz_state(nqubits: int, **kwargs):
    """Generate an :math:`n`-qubit Greenberger-Horne-Zeilinger (GHZ) state that takes the form

    .. math::
        \\ket{\\text{GHZ}} = \\frac{\\ket{0}^{\\otimes n} + \\ket{1}^{\\otimes n}}{\\sqrt{2}}

    where :math:`n` is the number of qubits.

    Args:
        nqubits (int): number of qubits :math:`n >= 2`.
        kwargs (dict, optional): additional arguments used to initialize a Circuit object.
            For details, see the documentation of :class:`qibo.models.circuit.Circuit`.

    Returns:
        :class:`qibo.models.circuit.Circuit`: Circuit that prepares the GHZ state.
    """
    if nqubits < 2:
        raise_error(
            ValueError,
            f"nqubits given as {nqubits}. nqubits needs to be >= 2.",
        )

    circuit = Circuit(nqubits, **kwargs)
    circuit.add(gates.H(0))
    circuit.add(gates.CNOT(qubit, qubit + 1) for qubit in range(nqubits - 1))

    return circuit


def graph_state(matrix, backend=None, **kwargs):
    """Create circuit encoding an undirected graph state given its adjacency matrix.

    Given a graph :math:`G = (V, E)` with :math:`V` being the set of vertices and :math:`E`
    being the set of edges, an $n$-qubit graph state is defined as

    .. math::
        \\ket{G} = \\prod_{(j, k) \\in E} \\, CZ_{j, k} \\, \\ket{+}^{\\otimes V} \\, ,

    where :math:`CZ_{a,b}` is the :class:`qibo.gates.CZ` gate acting on qubits :math:`j`
    and :math:`k`, and :math:`\\ket{+} = H \\, \\ket{0}$, with :class:`qibo.gates.H`
    being the Hadamard gate.

    Args:
        matrix (ndarray or list): Adjacency matrix of the graph.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend. Defaults to ``None``.
        kwargs (dict, optional): Additional arguments used to initialize a Circuit object.
            For details, see the documentation of :class:`qibo.models.circuit.Circuit`.

    Returns:
        :class:`qibo.models.circuit.Circuit`: Circuit of the graph state with the given
        adjacency matrix.
    """
    backend = _check_backend(backend)

    if isinstance(matrix, list):
        matrix = backend.cast(matrix, dtype=int)

    if not backend.np.allclose(matrix, matrix.T):
        raise_error(
            ValueError,
            f"``matrix`` is not symmetric, not representing an undirected graph",
        )

    nqubits = len(matrix)

    circuit = Circuit(nqubits, **kwargs)
    circuit.add(gates.H(qubit) for qubit in range(nqubits))

    # since the matrix is symmetric, we only need the upper triangular part
    rows, columns = backend.np.nonzero(backend.np.triu(matrix))
    circuit.add(gates.CZ(int(ind_r), int(ind_c)) for ind_r, ind_c in zip(rows, columns))

    return circuit


def dicke_state(nqubits: int, weight: int, all_to_all: bool = False, **kwargs):
    """Create a circuit that prepares the Dicke state :math:`\\ket{D_{k}^{n}}`.

    The Dicke state :math:`\\ket{D_{k}^{n}}` is the equal superposition of all :math:`n`-qubit
    computational basis states with fixed-Hamming-weight :math:`k`.
    The circuit prepares the state deterministically with :math:`O(k \\, n)` gates and :math:`O(n)`
    depth, or :math:`O(k \\log\\frac{n}{k})` depth under the assumption of ``all-to-all``
    connectivity.

    Args:
        nqubits (int): number of qubits :math:`n`.
        weight (int): Hamming weight :math:`k` of the Dicke state.
        all_to_all (bool, optional): If ``False``, uses implementation from Ref. [1].
            If ``True``, uses shorter-depth implementation from Ref. [2].
            Defaults to ``False``.
        kwargs (dict, optional): additional arguments used to initialize a Circuit object.
            For details, see the documentation of :class:`qibo.models.circuit.Circuit`.

    Returns:
        :class:`qibo.models.circuit.Circuit` : Circuit that prepares :math:`\\ket{D_{k}^{n}}`.

    References:
        1. Andreas Bärtschi and Stephan Eidenbenz, *Deterministic preparation of Dicke states*,
        `22nd International Symposium on Fundamentals of Computation Theory, FCT'19, 126-139  (2019)
        <https://doi.org/10.1007/978-3-030-25027-0_9>`_.

        2. Andreas Bärtschi and Stephan Eidenbenz, *Short-Depth Circuits for Dicke State Preparation*,
        `IEEE International Conference on Quantum Computing & Engineering (QCE), 87--96 (2022)
        <https://doi.org/10.1109/QCE53715.2022.00027>`_.
    """
    if weight < 0 or weight > nqubits:
        raise_error(
            ValueError, f"weight must be between 0 and {nqubits}, but got {weight}."
        )

    circuit = Circuit(nqubits, **kwargs)

    if weight == 0:
        return circuit

    if not all_to_all:
        # Start with |0⟩^(n-k) |1⟩^k
        circuit.add(gates.X(qubit) for qubit in range(nqubits - weight, nqubits))

        _add_dicke_unitary_gate(circuit, range(nqubits), weight)

        return circuit

    # We prepare disjoint sets of qubits
    disjoint_sets = [
        {
            "qubits": list(range(weight * it, weight * (it + 1))),
            "tier": weight,
            "children": [],
        }
        for it in range(nqubits // weight)
    ]
    nmodk = nqubits % weight
    if nmodk != 0:
        disjoint_sets.append(
            {
                "qubits": list(range(nqubits - nmodk, nqubits)),
                "tier": nmodk,
                "children": [],
            }
        )
    # reverse to have in ascending order of tier
    disjoint_sets = list(reversed(disjoint_sets))

    trees = disjoint_sets.copy()
    # combine lowest tier trees into one tree
    while len(trees) > 1:
        first_smallest = trees.pop(0)
        second_smallest = trees.pop(0)
        second_smallest["tier"] += first_smallest["tier"]
        second_smallest["children"].append(first_smallest)
        new = second_smallest
        # put new combined tree in list mantaining ordering
        trees.insert(sum(x["tier"] < new["tier"] for x in trees), new)

    root = trees[0]
    # We initialize |0>^(n-k)|1>^k  bitstring at root qubits
    circuit.add(gates.X(q) for q in root["qubits"][-weight:])

    # undo the union-by-tier algorithm:
    # split each root's (x) highest tier child y, and add WBD acting on both
    while len(trees) < len(disjoint_sets):
        cut_nodes = []
        for node in trees:
            if len(node["children"]) > 0:
                # find highest tier child
                y = max(node["children"], key=lambda x: x["tier"])
                # add WBD acting on both sets of qubits
                _add_wbd_gate(
                    circuit,
                    node["qubits"],
                    y["qubits"],
                    node["tier"],
                    y["tier"],
                    weight,
                )
                # cut tree splitting x and y
                node["tier"] -= y["tier"]
                node["children"].remove(y)
                cut_nodes.append(y)
        trees += cut_nodes

    for node in disjoint_sets:
        _add_dicke_unitary_gate(
            circuit, node["qubits"], min(weight, len(node["qubits"]))
        )

    return circuit


def _generate_rbs_pairs(nqubits: int, architecture: str, **kwargs):
    """Generate list of indexes representing the RBS connections.

    Creates circuit with all RBS initialised with 0.0 phase.

    Args:
        nqubits (int): number of qubits.
        architecture(str, optional): circuit architecture used for the unary loader.
            If ``diagonal``, uses a ladder-like structure.
            If ``tree``, uses a binary-tree-based structure.
            Defaults to ``tree``.
        kwargs (dict, optional): Additional arguments used to initialize a Circuit object.
            For details, see the documentation of :class:`qibo.models.circuit.Circuit`.

    Returns:
        (:class:`qibo.models.circuit.Circuit`, list): Circuit composed of
        :class:`qibo.gates.gates.RBS` and list of indexes of target qubits per depth.
    """

    if architecture == "diagonal":
        pairs_rbs = np.arange(nqubits)
        pairs_rbs = [[pair] for pair in zip(pairs_rbs[:-1], pairs_rbs[1:])]

    if architecture == "tree":
        pairs_rbs = [[(0, int(nqubits / 2))]]
        indexes = list(pairs_rbs[0][0])
        for depth in range(2, int(math.log2(nqubits)) + 1):
            pairs_rbs_per_depth = [
                [(index, index + int(nqubits / 2**depth)) for index in indexes]
            ]
            pairs_rbs += pairs_rbs_per_depth
            indexes = list(np.array(pairs_rbs_per_depth).flatten())

    pairs_rbs = [
        [(nqubits - 1 - a, nqubits - 1 - b) for a, b in row] for row in pairs_rbs
    ]

    circuit = Circuit(nqubits, **kwargs)
    for row in pairs_rbs:
        for pair in row:
            circuit.add(gates.RBS(*pair, 0.0, trainable=True))

    return circuit, pairs_rbs


def _generate_rbs_angles(data, architecture: str, nqubits: int = None, backend=None):
    """Generate list of angles for RBS gates based on ``architecture``.

    Args:
        data (ndarray, optional): :math:`1`-dimensional array of data to be loaded.
        architecture(str, optional): circuit architecture used for the unary loader.
            If ``diagonal``, uses a ladder-like structure.
            If ``tree``, uses a binary-tree-based structure.
            Defaults to ``tree``.
        nqubits (int): Number of qubits. To be used then ``architecture="tree"``.

    Returns:
        list: List of phases for RBS gates.
    """
    backend = _check_backend(backend)

    if architecture == "diagonal":
        phases = [
            backend.np.arctan2(backend.calculate_vector_norm(data[k + 1 :]), data[k])
            for k in range(len(data) - 2)
        ]
        phases.append(backend.np.arctan2(data[-1], data[-2]))

    if architecture == "tree":
        if nqubits is None:  # pragma: no cover
            raise_error(
                TypeError,
                '``nqubits`` must be specified when ``architecture=="tree"``.',
            )

        j_max = int(nqubits / 2)

        r_array = np.zeros(nqubits - 1, dtype=float)
        phases = np.zeros(nqubits - 1, dtype=float)
        for j in range(1, j_max + 1):
            r_array[j_max + j - 2] = math.sqrt(
                data[2 * j - 1] ** 2 + data[2 * j - 2] ** 2
            )
            theta = math.acos(data[2 * j - 2] / r_array[j_max + j - 2])
            if data[2 * j - 1] < 0.0:
                theta = 2 * math.pi - theta
            phases[j_max + j - 2] = theta

        for j in range(j_max - 1, 0, -1):
            r_array[j - 1] = math.sqrt(r_array[2 * j] ** 2 + r_array[2 * j - 1] ** 2)
            phases[j - 1] = math.acos(r_array[2 * j - 1] / r_array[j - 1])

    phases = backend.cast(phases, dtype=phases[0].dtype)

    return phases


def _parametrized_two_qubit_gate(gate, q0, q1, params=None):
    """Return two-qubit gate initialized with or without phases."""
    if params is not None:
        return gate(q0, q1, *params)

    return gate(q0, q1)


def _next_nearest_layer(
    nqubits: int, gate, parameters, closed_boundary: bool, **kwargs
):
    """Create entangling layer with next-nearest-neighbour connectivity."""
    circuit = Circuit(nqubits, **kwargs)
    circuit.add(
        _parametrized_two_qubit_gate(gate, qubit, qubit + 2, parameters)
        for qubit in range(nqubits - 2)
    )

    if closed_boundary:
        circuit.add(_parametrized_two_qubit_gate(gate, nqubits - 1, 0, parameters))

    return circuit


def _pyramid_layer(nqubits: int, gate, parameters, **kwargs):
    """Create entangling layer in triangular shape."""
    _, pairs_gates = _generate_rbs_pairs(nqubits, architecture="diagonal")
    pairs_gates = pairs_gates[::-1]

    circuit = Circuit(nqubits, **kwargs)
    circuit.add(
        _parametrized_two_qubit_gate(gate, pair[0][1], pair[0][0], parameters)
        for pair in pairs_gates
    )
    circuit.add(
        _parametrized_two_qubit_gate(gate, pair[0][1], pair[0][0], parameters)
        for k in range(1, len(pairs_gates))
        for pair in pairs_gates[:-k]
    )

    return circuit


def _v_layer(nqubits: int, gate, parameters, **kwargs):
    """Create entangling layer in V shape."""
    _, pairs_gates = _generate_rbs_pairs(nqubits, architecture="diagonal")
    pairs_gates = pairs_gates[::-1]

    circuit = Circuit(nqubits, **kwargs)
    circuit.add(
        _parametrized_two_qubit_gate(gate, pair[0][1], pair[0][0], parameters)
        for pair in pairs_gates
    )
    circuit.add(
        _parametrized_two_qubit_gate(gate, pair[0][1], pair[0][0], parameters)
        for pair in pairs_gates[::-1][1:]
    )

    return circuit


def _x_layer(nqubits, gate, parameters, **kwargs):
    """Create entangling layer in X shape."""
    _, pairs_gates = _generate_rbs_pairs(nqubits, architecture="diagonal")
    pairs_gates = pairs_gates[::-1]

    middle = int(np.floor(len(pairs_gates) / 2))
    pairs_1 = pairs_gates[:middle]
    pairs_2 = pairs_gates[-middle:]

    circuit = Circuit(nqubits, **kwargs)

    for first, second in zip(pairs_1, pairs_2[::-1]):
        circuit.add(
            _parametrized_two_qubit_gate(gate, first[0][1], first[0][0], parameters)
        )
        circuit.add(
            _parametrized_two_qubit_gate(gate, second[0][1], second[0][0], parameters)
        )

    circuit.add(
        _parametrized_two_qubit_gate(
            gate,
            pairs_gates[middle][0][1],
            pairs_gates[middle][0][0],
            parameters,
        )
    )

    for first, second in zip(pairs_1[::-1], pairs_2):
        circuit.add(
            _parametrized_two_qubit_gate(gate, first[0][1], first[0][0], parameters)
        )
        circuit.add(
            _parametrized_two_qubit_gate(gate, second[0][1], second[0][0], parameters)
        )

    return circuit


def _non_trivial_layers(
    nqubits: int,
    architecture: str = "pyramid",
    entangling_gate: Union[str, gates.Gate] = "RBS",
    closed_boundary: bool = False,
    **kwargs,
):
    """Create more intricate entangling layers of different shapes.

    Args:
        nqubits (int): number of qubits.
        architecture (str, optional): Architecture of the entangling layer.
            In alphabetical order, options are ``"next_nearest"``, ``"pyramid"``,
            ``"v"``, and ``"x"``. The ``"x"`` architecture is only defined for
            an even number of qubits. Defaults to ``"pyramid"``.
        entangling_gate (str or :class:`qibo.gates.Gate`, optional): Two-qubit gate to be used
            in the entangling layer. If ``entangling_gate`` is a parametrized gate,
            all phases are initialized as :math:`0.0`. Defaults to  ``"CNOT"``.
        closed_boundary (bool, optional): If ``True`` and ``architecture="next_nearest"``,
            adds a closed-boundary condition to the entangling layer. Defaults to ``False``.
        kwargs (dict, optional): Additional arguments used to initialize a Circuit object.
            For details, see the documentation of :class:`qibo.models.circuit.Circuit`.

    Returns:
        :class:`qibo.models.circuit.Circuit`: Circuit containing layer of two-qubit gates.
    """

    gate = (
        getattr(gates, entangling_gate)
        if isinstance(entangling_gate, str)
        else entangling_gate
    )

    parameters = list(signature(gate).parameters)
    parameters = (0.0,) * (len(parameters) - 3) if len(parameters) > 2 else None

    if architecture == "next_nearest":
        return _next_nearest_layer(nqubits, gate, parameters, closed_boundary, **kwargs)

    if architecture == "v":
        return _v_layer(nqubits, gate, parameters, **kwargs)

    if architecture == "x":
        return _x_layer(nqubits, gate, parameters, **kwargs)

    return _pyramid_layer(nqubits, gate, parameters, **kwargs)


def _angle_mod_two_pi(angle):
    """Return angle mod 2pi."""
    return angle % (2 * np.pi)


def _get_markers(bitstring, last_run: bool = False):
    """Subroutine of the Ehrlich algorithm."""
    nqubits = len(bitstring)
    markers = [len(bitstring) - 1]
    for ind, value in zip(range(nqubits - 2, -1, -1), bitstring[::-1][1:]):
        if value == bitstring[-1]:
            markers.append(ind)
        else:
            break

    markers = set(markers)

    if not last_run:
        markers = set(range(nqubits)) - markers

    return markers


def _get_next_bistring(bitstring, markers, hamming_weight):
    """Subroutine of the Ehrlich algorithm."""
    if len(markers) == 0:  # pragma: no cover
        return bitstring

    new_bitstring = np.copy(bitstring)

    nqubits = len(new_bitstring)

    indexes = np.argsort(bitstring)
    zeros, ones = np.sort(indexes[:-hamming_weight]), np.sort(indexes[-hamming_weight:])

    max_index = max(markers)
    nearest_one = ones[ones > max_index]
    nearest_one = None if len(nearest_one) == 0 else nearest_one[0]
    if new_bitstring[max_index] == 0 and nearest_one is not None:
        new_bitstring[max_index] = 1
        new_bitstring[nearest_one] = 0
    else:
        farthest_zero = zeros[zeros > max_index]
        if nearest_one is not None:
            farthest_zero = farthest_zero[farthest_zero < nearest_one]
        farthest_zero = farthest_zero[-1]
        new_bitstring[max_index] = 0
        new_bitstring[farthest_zero] = 1

    markers.remove(max_index)
    last_run = _get_markers(new_bitstring, last_run=True)
    markers = markers | (set(range(max_index + 1, nqubits)) - last_run)

    new_ones = np.argsort(new_bitstring)[-hamming_weight:]
    controls = list(set(ones) & set(new_ones))
    difference = new_bitstring - bitstring
    qubits = [np.where(difference == -1)[0][0], np.where(difference == 1)[0][0]]

    return new_bitstring, markers, [qubits, controls]


def _ehrlich_algorithm(initial_string, return_indices: bool = True):
    """Return list of bitstrings with mininal Hamming distance between consecutive strings.

    Based on the Gray code called Ehrlich algorithm. For more details, please see Ref. [1].

    Args:
        initial_string (ndarray): initial bitstring as an :math:`1`-dimensional array
            of size :math:`n`. All ones in the bitstring need to be consecutive.
            For instance, for :math:`n = 6` and :math:`k = 2`, the bistrings
            :math:`000011` and :math:`001100` are examples of acceptable inputs.
            In contrast, :math:`001001` is not an acceptable input.
        return_indices (bool, optional): if ``True``, returns the list of indices of
            qubits that act like controls and targets of the circuit to be created.
            Defaults to ``True``.

    Returns:
        list or tuple(list, list): If ``return_indices=False``, returns list containing
        sequence of bistrings in the order generated by the Gray code.
        If ``return_indices=True`` returns tuple with the aforementioned list and the list
        of control anf target qubits of gates to be implemented based on the sequence of
        bitstrings.

    References:
        1. R. M. S. Farias, T. O. Maciel, G. Camilo, R. Lin, S. Ramos-Calderer, and L. Aolita,
        *Quantum encoder for fixed-Hamming-weight subspaces*
        `Phys. Rev. Applied 23, 044014 (2025) <https://doi.org/10.1103/PhysRevApplied.23.044014>`_.
    """
    k = np.unique(initial_string, return_counts=True)
    if len(k[1]) == 1:  # pragma: no cover
        return ["".join([str(item) for item in initial_string])]

    k = k[1][1]
    n = len(initial_string)
    n_choose_k = int(binom(n, k))

    markers = _get_markers(initial_string, last_run=False)
    string = initial_string
    strings = ["".join(string[::-1].astype(str))]
    controls_and_targets = []
    for _ in range(n_choose_k - 1):
        string, markers, c_and_t = _get_next_bistring(string, markers, k)
        strings.append("".join(string[::-1].astype(str)))
        controls_and_targets.append(c_and_t)

    if return_indices:
        return strings, controls_and_targets

    return strings


def _get_gate(
    qubits_in: List[int],
    qubits_out: List[int],
    controls: List[int],
    theta: float,
    phi: float,
    complex_data: bool = False,
):
    """Return gate(s) necessary to encode a complex amplitude in a given computational basis state,
    given the computational basis state used to encode the previous amplitude.

    Information about computational basis states in question are contained
    in the indexing of ``qubits_in``, ``qubits_out``, and ``controls``.

    Args:
        qubits_in (list): list of qubits with ``in`` label.
        qubits_out (list): list of qubits with ``out`` label.
        controls (list): list of qubits that control the resulting gate.
        theta (float): first phase, used to encode the ``abs`` of amplitude.
        phi (float): second phase, used to encode the complex phase of amplitude.
        complex_data (bool): if ``True``, uses :class:`qibo.gates.U3` to as basis gate.
            If ``False``, uses :class:`qibo.gates.RY` as basis gate. Defaults to ``False``.

    Returns:
        List[:class:`qibo.gates.Gate`]: gate(s) to be added to circuit.

    References:
        1. R. M. S. Farias, T. O. Maciel, G. Camilo, R. Lin, S. Ramos-Calderer, and L. Aolita,
        *Quantum encoder for fixed-Hamming-weight subspaces*
        `Phys. Rev. Applied 23, 044014 (2025) <https://doi.org/10.1103/PhysRevApplied.23.044014>`_.
    """
    if len(qubits_in) == 0 and len(qubits_out) == 1:  # pragma: no cover
        gate_list = (
            gates.U3(*qubits_out, 2 * theta, 2 * phi, 0.0).controlled_by(*controls)
            if complex_data
            else gates.RY(*qubits_out, 2 * theta).controlled_by(*controls)
        )
        gate_list = [gate_list]
    elif len(qubits_in) == 1 and len(qubits_out) == 1:
        ## chooses best combination of complex RBS gate
        ## given number of controls and if data is real or complex
        gate_list = []
        gate = gates.RBS(*qubits_in, *qubits_out, theta)
        if len(controls) > 0:
            gate = gate.controlled_by(*controls)
        gate_list.append(gate)

        if complex_data:
            gate = [gates.RZ(*qubits_in, -phi), gates.RZ(*qubits_out, phi)]
            if len(controls) > 0:
                gate = [g.controlled_by(*controls) for g in gate]
            gate_list.extend(gate)

    else:  # pragma: no cover
        # Important for future sparse encoder
        gate_list = [
            gates.GeneralizedRBS(
                list(qubits_in), list(qubits_out), theta, phi
            ).controlled_by(*controls)
        ]

    return gate_list


def _get_phase_gate_correction(last_string, phase: float):
    """Return final gate of HW-k circuits that encode complex data."""

    # to avoid circular import error
    from qibo.quantum_info.utils import (  # pylint: disable=import-outside-toplevel
        hamming_weight,
    )

    if isinstance(last_string, str):
        last_string = np.asarray(list(last_string), dtype=int)

    last_weight = hamming_weight(last_string)
    last_ones = np.argsort(last_string)
    last_zero = int(last_ones[0])
    last_controls = [int(qubit) for qubit in last_ones[-last_weight:]]

    # adding an RZ gate to correct the phase of the last amplitude encoded
    return gates.RZ(last_zero, 2 * phase).controlled_by(*last_controls)


def _binary_encoder_hopf(
    data, nqubits, complex_data, backend=None, **kwargs
):  # pylint: disable=unused-argument
    # TODO: generalize to complex-valued data
    backend = _check_backend(backend)

    dims = 2**nqubits

    base_strings = [f"{elem:0{nqubits}b}" for elem in range(dims)]
    base_strings = backend.np.reshape(base_strings, (-1, 2))
    strings = [base_strings]
    for _ in range(nqubits - 1):
        base_strings = backend.np.reshape(base_strings[:, 0], (-1, 2))
        strings.append(base_strings)
    strings = strings[::-1]

    targets_and_controls = []
    for pairs in strings:
        for pair in pairs:
            targets, controls, anticontrols = [], [], []
            for k, bits in enumerate(zip(pair[0], pair[1])):
                if bits == ("0", "0"):
                    anticontrols.append(k)
                elif bits == ("1", "1"):
                    controls.append(k)
                elif bits == ("0", "1"):
                    targets.append(k)
            targets_and_controls.append([targets, controls, anticontrols])

    circuit = Circuit(nqubits, **kwargs)
    for targets, controls, anticontrols in targets_and_controls:
        gate_list = []
        if len(anticontrols) > 0:
            gate_list.append(gates.X(qubit) for qubit in anticontrols)
        gate_list.append(
            gates.RY(targets[0], 0.0).controlled_by(*(controls + anticontrols))
        )
        if len(anticontrols) > 0:
            gate_list.append(gates.X(qubit) for qubit in anticontrols)
        circuit.add(gate_list)

    angles = _generate_rbs_angles(data, "tree", dims, backend=backend)
    circuit.set_parameters(2 * angles)

    return circuit


def _binary_encoder_hyperspherical(
    data, nqubits, complex_data: bool, backend=None, **kwargs
):
    backend = _check_backend(backend)

    dims = 2**nqubits
    last_qubit = nqubits - 1

    indexes_to_double, lex_order_global = [0], [0]

    circuit = Circuit(nqubits, **kwargs)
    if complex_data:
        circuit.add(gates.U3(last_qubit, 0.0, 0.0, 0.0))
    else:
        circuit.add(gates.RY(last_qubit, 0.0))

    cummul_n_k = 0
    initial_string = np.array([1] + [0] * (nqubits - 1))
    for weight in range(1, nqubits):
        n_choose_k = int(binom(nqubits, weight))
        cummul_n_k += n_choose_k
        placeholder = np.random.rand(n_choose_k)
        if complex_data:
            placeholder = placeholder.astype(complex) + 1j * np.random.rand(n_choose_k)
        placeholder = backend.cast(placeholder, dtype=placeholder[0].dtype)

        circuit += hamming_weight_encoder(
            placeholder,
            nqubits,
            weight,
            full_hwp=True,
            optimize_controls=False,
            phase_correction=False,
            initial_string=initial_string,
            backend=backend,
            **kwargs,
        )

        # add gate to be place between blocks of Hamming-weight encoders
        gate, lex_order, initial_string, phase_index = _intermediate_gate(
            initial_string,
            weight,
            last_qubit,
            cummul_n_k,
            complex_data,
        )
        circuit.add(gate)
        lex_order_global.extend(lex_order)
        indexes_to_double.append(phase_index)

    # sort data such that the encoding is performed in lexicographical order
    lex_order_global.append(dims - 1)
    lex_order_sorted = np.sort(np.copy(lex_order_global))
    lex_order_global = [
        np.where(lex_order_sorted == num)[0][0] for num in lex_order_global
    ]
    data = data[lex_order_global]
    del lex_order_global, lex_order_sorted

    _data = backend.np.abs(data) if complex_data else data

    thetas = _generate_rbs_angles(_data, architecture="diagonal", backend=backend)

    phis = backend.np.zeros(len(thetas) + 1)
    if complex_data:
        phis[0] = _angle_mod_two_pi(-backend.np.angle(data[0]))
        for k in range(1, len(phis)):
            phis[k] = _angle_mod_two_pi(
                -backend.np.angle(data[k]) + backend.np.sum(phis[:k])
            )
    phis = backend.cast(phis, dtype=phis[0].dtype)

    zero_casted = backend.cast(0.0, dtype=backend.np.float64)  # because of GPU backends

    angles = []
    for k, (theta, phi) in enumerate(zip(thetas, phis)):
        if k in indexes_to_double:
            angle = [2 * theta, 2 * phi, zero_casted] if complex_data else [2 * theta]
        else:
            angle = [theta, -phi, phi] if complex_data else [theta]

        angles.extend(angle)

    if complex_data:
        angles[-2] = 2 * _angle_mod_two_pi(
            (backend.np.angle(data[-1]) - backend.np.angle(data[-2])) / 2
        )
        angles[-1] = 2 * _angle_mod_two_pi(
            (-0.5) * (backend.np.angle(data[-2]) + backend.np.angle(data[-1]))
            + backend.np.sum(phis[:-2])
        )

    circuit.set_parameters(angles)

    return circuit


def _intermediate_gate(
    initial_string,
    weight,
    last_qubit,
    cummul_n_k,
    complex_data,
):
    """Calculate where to place the intermediate gate by finding the last string
    of the previous Hamming-weight block that was encoded"""

    # sort data such that the encoding is performed in lexicographical order
    bitstrings = _ehrlich_algorithm(initial_string, False)
    initial_string = bitstrings[-1]
    lex_order = [int(string, 2) for string in bitstrings]

    controls = [item.start() for item in finditer("1", initial_string)]
    index = (
        initial_string.find("0")
        if weight % 2 == 0
        else last_qubit - initial_string[::-1].find("0")
    )
    initial_string = np.array(list(initial_string), dtype=int)
    initial_string[index] = 1
    initial_string = initial_string[::-1]

    phase_index = cummul_n_k
    gate = (
        gates.U3(index, 0.0, 0.0, 0.0).controlled_by(*controls)
        if complex_data
        else gates.RY(index, 0.0).controlled_by(*controls)
    )

    return gate, lex_order, initial_string, phase_index


def _sort_data_sparse(data, nqubits, backend):
    from qibo.quantum_info.utils import (  # pylint: disable=import-outside-toplevel
        hamming_weight,
    )

    # TODO: Fix this mess with qibo native data types
    try:
        test_dtype = bool("int" in str(data[0][0].dtype))
    except AttributeError:
        test_dtype = bool("int" in str(type(data[0][0])))

    _data = [(f"{row[0]:0{nqubits}b}", row[1]) for row in data] if test_dtype else data

    _data = sorted(_data, key=lambda x: hamming_weight(x[0]))

    bitstrings_sorted, data_sorted = zip(*_data)
    bitstrings_sorted = [
        np.array(list(string)).astype(int) for string in bitstrings_sorted
    ]

    bitstrings_sorted = backend.cast(bitstrings_sorted, dtype=backend.np.int8)
    data_sorted = backend.cast(data_sorted, dtype=data_sorted[0].dtype)

    return data_sorted, bitstrings_sorted


def _get_gate_sparse(
    distance,
    difference,
    touched_qubits,
    complex_data,
    controls,
    hw_0,
    hw_1,
    theta,
    phi,
    backend=None,
):
    backend = _check_backend(backend)
    if distance == 1:
        qubit = int(backend.np.where(difference == 1)[0][0])
        if qubit not in touched_qubits:
            touched_qubits.append(qubit)
        gate = (
            gates.U3(qubit, 2 * theta, 2 * phi, 0.0).controlled_by(*controls)
            if complex_data
            else gates.RY(qubit, 2 * theta).controlled_by(*controls)
        )
    elif distance == 2 and hw_0 == hw_1:
        qubits = [
            int(np.where(difference == -1)[0][0]),
            int(np.where(difference == 1)[0][0]),
        ]
        touched_qubits += list(set(qubits) - set(touched_qubits))
        qubits_in = [int(qubits[0])]
        qubits_out = [int(qubits[1])]
        qubits = [np.where(difference == -1)[0][0], np.where(difference == 1)[0][0]]

        gate = _get_gate(
            qubits_in,
            qubits_out,
            controls,
            theta,
            phi,
            complex_data,
        )
    else:
        qubits = [np.where(difference == -1)[0], np.where(difference == 1)[0]]
        for row in qubits:
            row = [int(elem) for elem in row]
            touched_qubits += list(set(row) - set(touched_qubits))
        qubits_in = [int(qubit) for qubit in qubits[0]]
        qubits_out = [int(qubit) for qubit in qubits[1]]
        gate = gates.GeneralizedRBS(qubits_in, qubits_out, theta, -phi).controlled_by(
            *controls
        )

    return gate


def _get_phase_gate_correction_sparse(
    last_string,
    second_to_last_string,
    nqubits,
    last_data,
    second_to_last_data,
    circuit,
    phis,
):
    from qibo.quantum_info.utils import (  # pylint: disable=import-outside-toplevel
        hamming_weight,
    )

    hw_0 = hamming_weight(second_to_last_string)
    hw_1 = hamming_weight(last_string)
    if hw_1 == nqubits and hw_0 == nqubits - 1:
        phi = _angle_mod_two_pi(np.angle(last_data) - np.angle(second_to_last_data))
        lamb = _angle_mod_two_pi(
            -(np.angle(second_to_last_data) + np.angle(last_data))
            + 2 * np.sum(phis[:-2])
        )
        _gate = circuit.queue[-1]
        gate = gates.U3(
            *_gate.target_qubits, _gate.init_kwargs["theta"], phi, lamb
        ).controlled_by(*_gate.control_qubits)

        new_queue = circuit.queue[:-1] + [gate]

        return new_queue

    if hw_1 == nqubits:  # pragma: no cover
        first_one = np.argsort(second_to_last_string)[0]
        other_ones = list(set(range(nqubits)) ^ {first_one})
        gate = gates.RZ(first_one, 2 * phis[-1]).controlled_by(*other_ones)
    else:
        gate = _get_phase_gate_correction(last_string, phis[-1])

    return gate


def _add_scs_gate(circuit: Circuit, qubits: List[int], weight: int):
    """In-place addition of a Split & Cyclic Shift (SCS) gate to ``circuit``.
    Implements the SCS_{n,k} unitary from Definition 3 of the paper [1].
    Acts on the last weight+1 qubits of the nqubits passed qubits.
    """
    if weight == 0:
        return  # SCS_{n,0} is identity

    nqubits = len(qubits)
    last_qubit = qubits[-1]  # qubits[nqubits - 1]
    target_qubit = qubits[-2]
    # first_qubit = qubits[nqubits - 1 - weight]

    # Gate (i) - acts on last two qubits
    theta = 2 * np.arccos(np.sqrt(1 / nqubits))
    circuit.add(gates.CNOT(target_qubit, last_qubit))
    circuit.add(gates.RY(target_qubit, theta).controlled_by(last_qubit))
    circuit.add(gates.CNOT(target_qubit, last_qubit))

    # Gates (ii)_ℓ for ℓ from 2 to k
    for l in range(2, weight + 1):
        theta = 2 * np.arccos(np.sqrt(l / nqubits))
        target_qubit = qubits[-(l + 1)]
        control_qubit = qubits[-l]

        # Implement the three-qubit gate (ii)_ℓ
        circuit.add(gates.CNOT(target_qubit, last_qubit))
        circuit.add(
            gates.RY(target_qubit, theta).controlled_by(control_qubit, last_qubit)
        )
        circuit.add(gates.CNOT(target_qubit, last_qubit))


def _add_dicke_unitary_gate(circuit: Circuit, qubits: List[int], weight: int):
    """In-place addition to ``circuit`` of a U_{n,k} gate from Definition 2 of the paper [1]."""
    nqubits = len(qubits)
    for m in range(nqubits, weight, -1):
        # Add SCS_{m,k} acting on last k+1 qubits
        _add_scs_gate(circuit, qubits[:m], weight)

    # Recursively build the unitary U_n,k
    for m in range(weight, 0, -1):
        # Add SCS_{m,m-1} acting on last m qubits
        _add_scs_gate(circuit, qubits[:m], m - 1)


def _add_wbd_gate(
    circuit: Circuit,
    first_register: List[int],
    second_register: List[int],
    nqubits: int,
    mqubits: int,
    weight: int,
):
    """In-place addition of a Weight Distribution Block (WBD) to ``circuit``.
    Implements the :math:`WBD^{n,m}_k` unitary from Definition 2 of the paper [2].
    Only acts on first_register and second_register, last k qubits
    Our circuit is mirrored, as paper [2] uses a top-bottom circuit <-> right-left bitstring convention
    """

    if mqubits > nqubits / 2:
        raise_error(ValueError, "``m`` must not be greater than ``n - m``.")

    # only acts on last k qubits
    first_register = first_register[-weight:]
    second_register = second_register[-weight:]

    # if m>k, m is truncated. Operations involving the most significant k-m digits can be removed

    # (1) Switching from unary encoding to one hot encoding
    circuit.add(gates.CNOT(q, q + 1) for q in first_register[-2::-1])

    # (2) Adding a supperposition of hamming weights into the second register
    # this can be optimized
    # We follow Figure 4, but adjust definition of xi and si (suffix sum) to match
    theta_gate = lambda qubit, theta: gates.RY(qubit, 2 * math.acos(theta))
    for l in range(weight, 0, -1):
        x = [
            math.comb(mqubits, i) * math.comb(nqubits - mqubits, l - i)
            for i in range(l)
        ]
        s = math.comb(nqubits, l)
        circuit.add(
            theta_gate(second_register[-1], math.sqrt(x[0] / s)).controlled_by(
                first_register[-l]
            )
        )
        s -= x[0]
        for qubit in range(1, min(l, mqubits)):
            circuit.add(
                theta_gate(
                    second_register[-qubit - 1], math.sqrt(x[qubit] / s)
                ).controlled_by(first_register[-l], second_register[-qubit])
            )
            s -= x[qubit]

    # (3) Go back to unary encoding, undoing one hot encoding
    circuit.add(gates.CNOT(q, q + 1) for q in first_register[:-1])

    # (4) Substracting weight i in the first register from weight l in the second register.
    # fredkin, controlled swaps (decomposed into CNOT and Toffoli)
    fredkin = lambda control, s1, s2: (
        gates.CNOT(s2, s1),
        gates.TOFFOLI(control, s1, s2),
        gates.CNOT(s2, s1),
    )

    dif = max(0, weight - mqubits)
    for control in range(dif, weight):
        for q in range(control):
            circuit.add(
                fredkin(
                    second_register[control - dif],
                    first_register[control - q],
                    first_register[control - q - 1],
                )
            )
        circuit.add(gates.CNOT(second_register[control - dif], first_register[0]))


def _perm_column_ops(
    indices: list[int],
    n: int,
    backend=None,
):
    """Return (ell, gate_list) performing duplicate‑col removal + compaction."""
    backend = _check_backend(backend)
    # flatten the (x_0, x_1),...(x_{2m-2}, x_{2m-1})
    # We construct a matrix composed of
    # xj to track the changes in xj where xj,k represents the k-th bit and the
    # bits are arranged from the least significant bit to the most significant bit
    indices = list(sum(indices, ()))
    A = []
    for x in indices:
        bits = []
        for k in range(n):
            bits.append((x >> k) & 1)
        A.append(bits)
    A = backend.cast(A, dtype=backend.np.int8)
    ncols = A.shape[1]
    # initialize the list of gates
    qgates = []

    # number of non-zero columns
    ell = 0
    flag = backend.np.zeros(n, dtype=int)
    for idxj in range(ncols):
        if any(elem != 0 for elem in A[:, idxj]):
            ell += 1
            flag[idxj] = 1

            # look for columns that are equal to A[:,idxj]
            for idxk in range(idxj + 1, ncols):
                if backend.np.array_equal(A[:, idxj], A[:, idxk]):
                    qgates.append(gates.CNOT(n - idxj - 1, n - idxk - 1))
                    # this should transform the k-th column into an all-zero column
                    A[:, idxk] = 0

    # Now, we need to swap the ell non-zero columns to the first ell columns
    for idxk in range(ell, ncols):
        if not backend.np.array_equal(A[:, idxk], backend.np.zeros_like(A[:, idxk])):
            for k in range(len(flag)):
                if flag[k] == 0:
                    flag[k] = 1
                    flag[idxk] = 0

                    qgates.append(gates.SWAP(n - idxk - 1, n - k - 1))

                    bits = A[:, idxk].copy()
                    A[:, idxk] = A[:, k]
                    A[:, k] = bits
                    break

    return ell, qgates, A


def _perm_row_ops(A, ell: int, m: int, n: int, backend=None):
    """Return gates that reduce all rows after row0 to target form."""
    backend = _check_backend(backend)

    log2m = int(backend.np.log2(2 * m))
    atilde = backend.np.array(
        [[(x >> k) & 1 for k in range(n)] for x in range(2 * m)], dtype=int
    )

    qgates = []
    nrows = A.shape[0]
    ncols = A.shape[1]
    # Start with the first row (indexed as row 0)
    for k in range(ncols):
        # If we find a0,k = 1 for any 0 <= k <= n−1
        if A[0, k] == 1:
            qgates.append(gates.X(n - k - 1))
            A[:, k] = (A[:, k] + 1) % 2

    for j in range(1, nrows):
        flag = False
        for k in range(log2m, ncols):
            if A[j, k] != 0:
                flag = True
                break

        if not flag:
            # There is no element b_{j},k != 0 for k > {log2m-1}"
            ctrls = [n - l - 1 for l in range(ncols) if A[j, l] != 0]
            qgates.append(gates.X(n - log2m - 1).controlled_by(*ctrls))
            ctrls = [l for l in range(ncols) if A[j, l] == 1]

            # check whether the gate is applied on other rows
            for l in range(j, nrows):
                if all(elem == 1 for elem in A[l, ctrls]):
                    A[l, log2m] = (A[l, log2m] + 1) % 2

        # There is always an element b_{j},k != 0 for k > {log2m-1}
        for k in range(log2m, ncols):
            if A[j, k] != 0:
                # Element b_{j},{k} != 0, {k} > {log2m-1}"
                for kprime in range(ell):
                    # There is a typo in the paper
                    # b_{j},{kprime} should be different from Ã_{j},{kprime} not Ã_{j},{k}
                    if kprime != k and A[j, kprime] != atilde[j, kprime]:
                        qgates.append(gates.CNOT(n - k - 1, n - kprime - 1))
                        # check whether the gate is applied on other rows
                        for l in range(nrows):
                            if A[l, k] == 1:
                                A[l, kprime] = (A[l, kprime] + 1) % 2

                # Let us clean the element b_{j},{k}

                # There is another typo in the paper
                # the control qubits for this gate correspond to the non-zero elements in row j of matrix A, not Ã
                ctrls = [n - l - 1 for l in range(k) if A[j, l] != 0]
                qgates.append(gates.X(n - k - 1).controlled_by(*ctrls))
                ctrls = [l for l in range(k) if A[j, l] != 0]

                # check whether the gate is applied on other rows
                for l in range(nrows):
                    if all(elem == 1 for elem in A[l, ctrls]):
                        A[l, k] = (A[l, k] + 1) % 2

    return qgates


def _perm_pair_flip_ops(n: int, m: int, backend=None):
    """Implement σ_{i,2} as X fan‑in + MCX + X fan‑out."""
    backend = _check_backend(backend)
    # let us flip the first qubit when the last {int(n-math.log2(2*m))} qubits are all in the state |0⟩
    prefix = int(backend.np.ceil(backend.np.log2(2 * m)))
    x_qubits, controls = range(prefix, n), range(n - prefix)
    qgates = [gates.X(n - q - 1) for q in x_qubits]
    qgates.append(gates.X(n - 1).controlled_by(*controls))  # flip qubit 0
    qgates.extend(gates.X(n - q - 1) for q in x_qubits)

    return qgates


def permutation_synthesis(
    sigma: Union[List[int], tuple[int, ...]], m: int = 2, backend=None, **kwargs
):
    """Return circuit that implements a given permutation.

    Given permutation ``sigma`` on :math:`\\{0, \\, 1, \\, \\dots, \\, d-1\\}`
    and a power‑of‑two budget ``m``, this function factors ``sigma``
    into the fewest layers :math:`\\sigma_{1}, \\, \\sigma_{2}, \\, \\cdots, \\, \\sigma_{t}`
    such that:
        - each layer has at most :math:`m` disjoint transpositions;
        - each layer moves a power‑of‑two number of indices.

    The function returns a circuit synthesis of ``sigma``.

    Args:
        sigma (list or tuple): permutation description on :math:`\\{0, \\, 1, \\, \\dots, \\, d-1\\}`.
        m (int): power‑of‑two budget. Defauls to :math:`2`.

    Returns:
        :class:`qibo.models.circuit.Circuit`: Circuit that implements the permutation ``sigma``.

    References:
        1. L. Li, and J. Luo,
        *Nearly Optimal Circuit Size for Sparse Quantum State Preparation*
        `arXiv:2406.16142 (2024) <https://doi.org/10.48550/arXiv.2406.16142>`_.
    """
    backend = _check_backend(backend)

    if isinstance(sigma, tuple):
        sigma = list(sigma)

    if not isinstance(sigma, (list, tuple)):
        raise_error(
            TypeError,
            f"Permutation ``sigma`` must be either a ``list`` or a ``tuple`` of ``int``s.",
        )

    nqubits = int(backend.np.ceil(backend.np.log2(len(sigma))))
    if sum([abs(s - i) for s, i in zip(sorted(sigma), range(2**nqubits))]) != 0:
        raise_error(
            ValueError, "Permutation sigma must contain all indices {0,...,n-1}"
        )

    if m > 0 and (m & (m - 1)) != 0:
        raise_error(ValueError, f"budget m must be a power‑of‑two")

    from qibo.quantum_info.utils import (  # pylint: disable=import-outside-toplevel
        decompose_permutation,
    )

    # factor sigma into the fewest layers such that
    # each layer has at most m disjoint transpositions, and
    # each layer moves a power‑of‑two number of indices
    layers = decompose_permutation(sigma, m)

    circuit = Circuit(nqubits)
    # in case we have more than one permutation to do, do it in layers
    for layer in layers:
        m = len(layer)
        ell, col_gates, A = _perm_column_ops(layer, nqubits, backend)
        row_gates = _perm_row_ops(A, ell, m, nqubits, backend)
        flip_gates = _perm_pair_flip_ops(nqubits, m, backend)
        col_row = col_gates + row_gates
        circuit.add(col_row)
        circuit.add(flip_gates)
        circuit.add(col_row[::-1])

    return circuit



"""Utility functions for the Quantum Information module."""

from functools import cache, reduce
from itertools import permutations
from math import factorial
from re import finditer
from typing import List, Optional, Tuple, Union

import numpy as np

from qibo.backends import Backend, _check_backend
from qibo.config import PRECISION_TOL, raise_error


@cache
def _get_single_paulis(order: str, backend: Backend):
    pauli_labels = {"I": backend.matrices.I()}
    pauli_labels.update(
        {label: getattr(backend.matrices, label) for label in ("X", "Y", "Z")}
    )
    return [pauli_labels[label] for label in order]


@cache
def _pauli_basis_normalization(nqubits: int):
    return float(np.sqrt(2**nqubits))


def hamming_weight(
    bitstring: Union[int, str, list, tuple], return_indexes: bool = False
):
    """Calculates the Hamming weight of a bitstring.

    The Hamming weight of a bistring is the number of :math:'1's that the bistring contains.

    Args:
        bitstring (int or str or tuple or list): bitstring to calculate the
            weight, either in binary or integer representation.
        return_indexes (bool, optional): If ``True``, returns the indexes of the
            non-zero elements. Defaults to ``False``.

    Returns:
        int or list: Hamming weight of bitstring or list of indexes of non-zero elements.
    """
    if not isinstance(return_indexes, bool):
        raise_error(
            TypeError,
            f"return_indexes must be type bool, but it is type {type(return_indexes)}",
        )

    bitstring = (
        "".join([str(bit) for bit in list(bitstring)])
        if not isinstance(bitstring, int)
        else f"{bitstring:b}"
    )

    indexes = [item.start() for item in finditer("1", bitstring)]

    if return_indexes:
        return indexes

    return len(indexes)


def hamming_distance(
    bitstring_1: Union[int, str, list, tuple],
    bitstring_2: Union[int, str, list, tuple],
    return_indexes: bool = False,
):
    """Calculates the Hamming distance between two bistrings.

    This is done by calculating the Hamming weight
    (:func:`qibo.quantum_info.utils.hamming_weight`) of ``| bitstring_1 - bitstring_2 |``.

    Args:
        bitstring_1 (int or str or list or tuple): fisrt bistring.
        bitstring_2 (int or str or list or tuple): second bitstring.
        return_indexes (bool, optional): If ``True``, returns the indexes of the
            non-zero elements. Defaults to ``False``.

    Returns:
        int or list: Hamming distance or list of indexes of non-zero elements.
    """
    if not isinstance(return_indexes, bool):
        raise_error(
            TypeError,
            f"return_indexes must be type bool, but it is type {type(return_indexes)}",
        )

    if isinstance(bitstring_1, int):
        bitstring_1 = f"{bitstring_1:b}"

    if isinstance(bitstring_2, int):
        bitstring_2 = f"{bitstring_2:b}"

    if not isinstance(bitstring_1, str):
        bitstring_1 = "".join([str(bit) for bit in bitstring_1])

    if not isinstance(bitstring_2, str):
        bitstring_2 = "".join([str(bit) for bit in bitstring_2])

    nbits = max(len(bitstring_1), len(bitstring_2))

    bitstring_1 = "0" * (nbits - len(bitstring_1)) + bitstring_1
    bitstring_2 = "0" * (nbits - len(bitstring_2)) + bitstring_2

    difference = np.array(list(bitstring_1), dtype=int) - np.array(
        list(bitstring_2), dtype=int
    )
    difference = np.abs(difference)
    difference = difference.astype(str)
    difference = "".join(difference)

    return hamming_weight(difference, return_indexes=return_indexes)


def hadamard_transform(array, implementation: str = "fast", backend=None):
    """Calculates the (fast) Hadamard Transform :math:`\\text{HT}` of a
    :math:`2^{n}`-dimensional vector or :math:`2^{n} \\times 2^{n}` matrix :math:`A`,
    where :math:`n` is the number of qubits in the system. If :math:`A` is a vector, then

    .. math::
        \\text{HT}(A) = \\frac{1}{2^{n / 2}} \\, H^{\\otimes n} \\, A \\,

    where :math:`H` is the :class:`qibo.gates.H` gate. If :math:`A` is a matrix, then

    .. math::
        \\text{HT}(A) = \\frac{1}{2^{n}} \\, H^{\\otimes n} \\, A \\, H^{\\otimes n} \\, .

    Args:
        array (ndarray): array or matrix.
        implementation (str, optional): if ``"regular"``, it uses the straightforward
            implementation of the algorithm with computational complexity of
            :math:`\\mathcal{O}(2^{2n})` for vectors and :math:`\\mathcal{O}(2^{3n})`
            for matrices. If ``"fast"``, computational complexity is
            :math:`\\mathcal{O}(n \\, 2^{n})` in both cases.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend.
            Defaults to ``None``.

    Returns:
        ndarray: (Fast) Hadamard Transform of ``array``.
    """
    backend = _check_backend(backend)

    if (
        len(array.shape) not in [1, 2]
        or (len(array.shape) == 1 and np.log2(array.shape[0]).is_integer() is False)
        or (
            len(array.shape) == 2
            and (
                np.log2(array.shape[0]).is_integer() is False
                or np.log2(array.shape[1]).is_integer() is False
            )
        )
    ):
        raise_error(
            TypeError,
            f"array must have shape (2**n,) or (2**n, 2**n), but it has shape {array.shape}.",
        )

    if isinstance(implementation, str) is False:
        raise_error(
            TypeError,
            f"implementation must be type str, but it is type {type(implementation)}.",
        )

    if implementation not in ["fast", "regular"]:
        raise_error(
            ValueError,
            f"implementation must be either `regular` or `fast`, but it is {implementation}.",
        )

    if implementation == "regular":
        nqubits = int(np.log2(array.shape[0]))
        hadamards = backend.np.real(
            reduce(backend.np.kron, [backend.matrices.H] * nqubits)
        )
        hadamards /= 2 ** (nqubits / 2)
        hadamards = backend.cast(hadamards, dtype=hadamards.dtype)

        array = hadamards @ array

        if len(array.shape) == 2:
            array = array @ hadamards

        return array

    array = _hadamard_transform_1d(array, backend=backend)

    if len(array.shape) == 2:
        array = _hadamard_transform_1d(array.T, backend=backend).T

    # needed for the tensorflow backend
    array = backend.cast(array, dtype=array.dtype)

    return array


def hellinger_distance(prob_dist_p, prob_dist_q, validate: bool = False, backend=None):
    """Calculates the Hellinger distance :math:`H` between two discrete probability distributions.

    For probabilities :math:`\\mathbf{p}` and :math:`\\mathbf{q}`, it is defined as

    .. math::
        H(\\mathbf{p} \\, , \\, \\mathbf{q}) = \\frac{1}{\\sqrt{2}} \\, \\|
            \\sqrt{\\mathbf{p}} - \\sqrt{\\mathbf{q}} \\|_{2}

    where :math:`\\|\\cdot\\|_{2}` is the Euclidean norm.

    Args:
        prob_dist_p (ndarray or list): discrete probability distribution :math:`p`.
        prob_dist_q (ndarray or list): discrete probability distribution :math:`q`.
        validate (bool, optional): If ``True``, checks if :math:`p` and :math:`q` are proper
            probability distributions. Defaults to ``False``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
            used in the execution. If ``None``, it uses
            the current backend. Defaults to ``None``.

    Returns:
        (float): Hellinger distance :math:`H(p, q)`.
    """
    backend = _check_backend(backend)

    if isinstance(prob_dist_p, list):
        prob_dist_p = backend.cast(prob_dist_p, dtype=np.float64)
    if isinstance(prob_dist_q, list):
        prob_dist_q = backend.cast(prob_dist_q, dtype=np.float64)

    if (len(prob_dist_p.shape) != 1) or (len(prob_dist_q.shape) != 1):
        raise_error(
            TypeError,
            "Probability arrays must have dims (k,) but have "
            + f"dims {prob_dist_p.shape} and {prob_dist_q.shape}.",
        )

    if (len(prob_dist_p) == 0) or (len(prob_dist_q) == 0):
        raise_error(TypeError, "At least one of the arrays is empty.")

    if validate:
        if (any(prob_dist_p < 0) or any(prob_dist_p > 1.0)) or (
            any(prob_dist_q < 0) or any(prob_dist_q > 1.0)
        ):
            raise_error(
                ValueError,
                "All elements of the probability array must be between 0. and 1..",
            )
        if backend.np.abs(backend.np.sum(prob_dist_p) - 1.0) > PRECISION_TOL:
            raise_error(ValueError, "First probability array must sum to 1.")

        if backend.np.abs(backend.np.sum(prob_dist_q) - 1.0) > PRECISION_TOL:
            raise_error(ValueError, "Second probability array must sum to 1.")

    distance = float(
        backend.calculate_vector_norm(
            backend.np.sqrt(prob_dist_p) - backend.np.sqrt(prob_dist_q)
        )
        / np.sqrt(2)
    )

    return distance


def hellinger_fidelity(prob_dist_p, prob_dist_q, validate: bool = False, backend=None):
    """Calculates the Hellinger fidelity between two discrete probability distributions.

    For probabilities :math:`p` and :math:`q`, the fidelity is defined as

    .. math::
        (1 - H^{2}(p, q))^{2} \\, ,

    where :math:`H(p, q)` is the :func:`qibo.quantum_info.hellinger_distance`.

    Args:
        prob_dist_p (ndarray or list): discrete probability distribution :math:`p`.
        prob_dist_q (ndarray or list): discrete probability distribution :math:`q`.
        validate (bool, optional): if ``True``, checks if :math:`p` and :math:`q` are proper
            probability distributions. Defaults to ``False``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
            used in the execution. If ``None``, it uses
            the current backend. Defaults to ``None``.

    Returns:
        float: Hellinger fidelity.

    """
    backend = _check_backend(backend)

    distance = hellinger_distance(prob_dist_p, prob_dist_q, validate, backend=backend)

    return (1 - distance**2) ** 2


def hellinger_shot_error(
    prob_dist_p, prob_dist_q, nshots: int, validate: bool = False, backend=None
):
    """Calculates the Hellinger fidelity error between two discrete probability distributions estimated from finite statistics.

    It is calculated propagating the probability error of each state of the system.
    The complete formula is:

    .. math::
        \\frac{1 - H^{2}(p, q)}{\\sqrt{nshots}} \\, \\sum_{k} \\,
            \\left(\\sqrt{p_{k} \\, (1 - q_{k})} + \\sqrt{q_{k} \\, (1 - p_{k})}\\right)

    where :math:`H(p, q)` is the :func:`qibo.quantum_info.hellinger_distance`,
    and :math:`1 - H^{2}(p, q)` is the square root of the
    :func:`qibo.quantum_info.hellinger_fidelity`.

    Args:
        prob_dist_p (ndarray or list): discrete probability distribution :math:`p`.
        prob_dist_q (ndarray or list): discrete probability distribution :math:`q`.
        nshots (int): number of shots we used to run the circuit to obtain :math:`p` and :math:`q`.
        validate (bool, optional): if ``True``, checks if :math:`p` and :math:`q` are proper
            probability distributions. Defaults to ``False``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
            used in the execution. If ``None``, it uses
            the current backend. Defaults to ``None``.

    Returns:
        float: Hellinger fidelity error.

    """
    backend = _check_backend(backend)

    if isinstance(prob_dist_p, list):
        prob_dist_p = backend.cast(prob_dist_p, dtype=np.float64)

    if isinstance(prob_dist_q, list):
        prob_dist_q = backend.cast(prob_dist_q, dtype=np.float64)

    hellinger_error = hellinger_fidelity(
        prob_dist_p, prob_dist_q, validate=validate, backend=backend
    )
    hellinger_error = np.sqrt(hellinger_error / nshots) * backend.np.sum(
        np.sqrt(prob_dist_q * (1 - prob_dist_p))
        + np.sqrt(prob_dist_p * (1 - prob_dist_q))
    )

    return hellinger_error


def total_variation_distance(
    prob_dist_p, prob_dist_q, validate: bool = False, backend=None
):
    """Calculate the total variation distance between two discrete probability distributions.

    For probabilities :math:`p` and :math:`q`, the total variation distance is defined as

    .. math::
        \\operatorname{TVD}(p, \\, q) = \\frac{1}{2} \\, \\|p - q\\|_{1}
            = \\frac{1}{2} \\, \\sum_{x} \\, \\left|p(x) - q(x)\\right| \\, ,

    where :math:`\\|\\cdot\\|_{1}` detones the :math:`\\ell_{1}`-norm.

    Args:
        prob_dist_p (ndarray or list): discrete probability distribution :math:`p`.
        prob_dist_q (ndarray or list): discrete probability distribution :math:`q`.
        validate (bool, optional): if ``True``, checks if :math:`p` and :math:`q` are proper
            probability distributions. Defaults to ``False``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
            used in the execution. If ``None``, it uses
            the current backend. Defaults to ``None``.

    Returns:
        float: Total variation distance measure.
    """
    backend = _check_backend(backend)

    if isinstance(prob_dist_p, list):
        prob_dist_p = backend.cast(prob_dist_p, dtype=np.float64)

    if isinstance(prob_dist_q, list):
        prob_dist_q = backend.cast(prob_dist_q, dtype=np.float64)

    if validate:
        if (any(prob_dist_p < 0) or any(prob_dist_p > 1.0)) or (
            any(prob_dist_q < 0) or any(prob_dist_q > 1.0)
        ):
            raise_error(
                ValueError,
                "All elements of the probability array must be between 0. and 1..",
            )
        if backend.np.abs(backend.np.sum(prob_dist_p) - 1.0) > PRECISION_TOL:
            raise_error(ValueError, "First probability array must sum to 1.")

        if backend.np.abs(backend.np.sum(prob_dist_q) - 1.0) > PRECISION_TOL:
            raise_error(ValueError, "Second probability array must sum to 1.")

    tvd = backend.calculate_vector_norm(prob_dist_p - prob_dist_q, order=1)

    return tvd / 2


def haar_integral(
    nqubits: int,
    power_t: int,
    samples: Optional[int] = None,
    backend=None,
):
    """Returns the integral over pure states over the Haar measure.

    .. math::
        \\int_{\\text{Haar}} d\\psi \\, \\left(|\\psi\\rangle\\right.\\left.
            \\langle\\psi|\\right)^{\\otimes t}

    Args:
        nqubits (int): Number of qubits.
        power_t (int): power that defines the :math:`t`-design.
        samples (int, optional): If ``None``, estimated the integral exactly.
            Otherwise, number of samples to estimate the integral via sampling.
            Defaults to ``None``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
            used in the execution. If ``None``, it uses
            the current backend. Defaults to ``None``.

    Returns:
        array: Estimation of the Haar integral.

    .. note::
        The ``exact=True`` method is implemented using Lemma 34 of
        `Kliesch and Roth (2020) <https://arxiv.org/abs/2010.05925>`_.
    """

    if isinstance(nqubits, int) is False:
        raise_error(
            TypeError, f"nqubits must be type int, but it is type {type(nqubits)}."
        )

    if isinstance(power_t, int) is False:
        raise_error(
            TypeError, f"power_t must be type int, but it is type {type(power_t)}."
        )

    if samples is not None and isinstance(samples, int) is False:
        raise_error(
            TypeError, f"samples must be type int, but it is type {type(samples)}."
        )

    backend = _check_backend(backend)

    dim = 2**nqubits

    if samples is not None:
        from qibo.quantum_info.random_ensembles import (  # pylint: disable=C0415
            random_statevector,
        )

        rand_unit_density = np.zeros((dim**power_t, dim**power_t), dtype=complex)
        rand_unit_density = backend.cast(
            rand_unit_density, dtype=rand_unit_density.dtype
        )

        random_states = backend.qinfo.ENGINE.random.standard_normal((samples, dim))
        random_states = backend.cast(random_states, dtype=rand_unit_density.dtype)
        random_states += 1.0j * backend.qinfo.ENGINE.random.standard_normal(
            (samples, dim)
        )
        random_states /= backend.qinfo.ENGINE.linalg.norm(
            random_states, axis=1
        ).reshape(-1, 1)
        random_states = random_states.reshape(samples, 1, dim)
        rho = backend.np.einsum(
            "ijk,ijl->ikl", random_states, backend.np.conj(random_states)
        )

        for state in rho:

            rand_unit_density = rand_unit_density + reduce(
                backend.np.kron, [state] * power_t
            )

        integral = rand_unit_density / samples

        return integral

    normalization = factorial(dim - 1) / factorial(dim - 1 + power_t)

    permutations_list = list(permutations(np.arange(power_t) + power_t))
    permutations_list = [
        tuple(np.arange(power_t)) + indices for indices in permutations_list
    ]

    identity = np.eye(dim**power_t, dtype=float)
    identity = np.reshape(identity, (dim,) * (2 * power_t))
    identity = backend.cast(identity, dtype=identity.dtype)

    integral = np.zeros((dim**power_t, dim**power_t), dtype=float)
    integral = backend.cast(integral, dtype=integral.dtype)
    for indices in permutations_list:
        integral = integral + backend.np.reshape(
            backend.np.transpose(identity, indices), (-1, dim**power_t)
        )
    integral = integral * normalization

    return integral


def pqc_integral(circuit, power_t: int, samples: int, backend=None):
    """Returns the integral over pure states generated by uniformly sampling
    in the parameter space described by a parameterized circuit.

    .. math::
        \\int_{\\Theta} d\\psi \\, \\left(|\\psi_{\\theta}\\rangle\\right.\\left.
            \\langle\\psi_{\\theta}|\\right)^{\\otimes t}

    Args:
        circuit (:class:`qibo.models.Circuit`): Parametrized circuit.
        power_t (int): power that defines the :math:`t`-design.
        samples (int): number of samples to estimate the integral.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
            used in the execution. If ``None``, it uses
            the current backend. Defaults to ``None``.

    Returns:
        ndarray: Estimation of the integral.
    """

    if isinstance(power_t, int) is False:
        raise_error(
            TypeError, f"power_t must be type int, but it is type {type(power_t)}."
        )

    if isinstance(samples, int) is False:
        raise_error(
            TypeError, f"samples must be type int, but it is type {type(samples)}."
        )

    backend = _check_backend(backend)

    circuit.density_matrix = True
    dim = 2**circuit.nqubits

    rand_unit_density = np.zeros((dim**power_t, dim**power_t), dtype=complex)
    rand_unit_density = backend.cast(rand_unit_density, dtype=rand_unit_density.dtype)
    for _ in range(samples):
        params = np.random.uniform(-np.pi, np.pi, circuit.trainable_gates.nparams)
        circuit.set_parameters(params)

        rho = backend.execute_circuit(circuit).state()

        rand_unit_density = rand_unit_density + reduce(np.kron, [rho] * power_t)

    integral = rand_unit_density / samples

    return integral


def _hadamard_transform_1d(array, backend=None):
    # necessary because of tf.EagerTensor
    # does not accept item assignment
    backend = _check_backend(backend)
    array_copied = backend.np.copy(array)

    indexes = [2**k for k in range(int(np.log2(len(array_copied))))]
    for index in indexes:
        for k in range(0, len(array_copied), 2 * index):
            for j in range(k, k + index):
                # copy necessary because of cupy backend
                elem_1 = backend.np.copy(array_copied[j])
                elem_2 = backend.np.copy(array_copied[j + index])
                array_copied[j] = elem_1 + elem_2
                array_copied[j + index] = elem_1 - elem_2
        array_copied /= 2.0

    return array_copied


def _cycles_from_perm(sigma: List[int]):
    """Extract the cycles from a permutation as follows:
    - Treat the permutation as a directed graph of arrows i->sigma(i).
    - Depth‑first walk from every unvisited vertex; each walk closes at the start -> a cycle.
    - Disjoint cycles partition the set {0,...,n-1} and commute, so we can factorize them independently.

    Args:
        sigma (list[int] or tuple[int]): permutation description on {0,...,n-1}.

    Returns:
        list[list[tuple[int, int]]]: :math:`t` list of disjoint cycles.
    """
    n, seen, cycles = len(sigma), [False] * len(sigma), []
    for i in range(n):
        # Depth‑first walk from every unvisited vertex
        if seen[i]:
            continue
        cur, cyc = i, []
        # Disjoint cycles
        while not seen[cur]:
            seen[cur] = True
            cyc.append(cur)
            cur = sigma[cur]
        if len(cyc) > 1:
            cycles.append(cyc)
    return cycles


def _star_matchings(cyc: list[int]):
    """
    Given a cycle :math:`(a_{1}, \\, a_{2}, \\, \\cdots, \\, a_{k})` with :math:`k \\geq 2`,
    the star factorization expresses the cycle as the ordered product of :math:`(k-1)`
    disjoint transpositions that all share the first vertex:

    .. math::
        (a_{1}, \\, a_{2}), \\, (a_{1}, \\, a_{3}), \\, \\cdots, \\, (a_{1}, \\, a_{k}) \\, .

    Applied right‑to‑left, this product reproduces the original cycle.

    Args:
        cyc (list[list[tuple[int, int]]]): :math:`t` list of disjoint cycles.

    Returns:
        list[list[tuple[int, int]]]: :math:`t` list of pairwise transpositions.
    """
    hub = cyc[0]
    return [[(min(hub, v), max(hub, v))] for v in cyc[1:]]


def _greedy_pack(matchings: List[List[Tuple[int, int]]], m: int):
    """
    Add a matching to the current layer if
        - it shares no vertex with swaps already in the layer, and
        - new layer size #swaps stays a power of two <= m.
    Otherwise flush the layer and start a new one.
    It works since disjointness keeps swaps commutative, and the power‑of‑two size rule
    aligns exactly with layer constraints.

    Args:
        matchings (list[list[tuple[int, int]]]): :math:`t` list of pairwise transpositions.

    Returns:
        list[list[tuple[int, int]]]: :math:`t` layers of pairwise transpositions.
    """
    layers: list[list[tuple[int, int]]] = []
    cur: list[tuple[int, int]] = []
    used = set()

    def _flush():
        nonlocal cur, used
        if cur:
            layers.append(cur)
        cur, used = [], set()

    def _verts(layer: list[tuple[int, int]]):
        for a, b in layer:
            yield a
            yield b

    for M in matchings:
        x = len(cur) + len(M)
        # is power of 2 and x<=m
        legal = (x > 0 and (x & (x - 1)) == 0) and x <= m
        # shares no vertex with swaps already in the layer
        if cur and (any(v in used for v in _verts(M)) or not legal):  # pragma: no cover
            _flush()
        cur.extend(M)
        used.update(_verts(M))
        if len(cur) == m:
            _flush()
    _flush()
    return layers


def decompose_permutation(
    sigma: Union[List[int], Tuple[int, ...]], m: int, backend=None
):
    """
     Given permutation ``sigma`` on :math:`\\{0, \\, 1, \\, \\dots, \\, d-1\\}`
    and a power‑of‑two budget ``m``, this function factors ``sigma``
    into the fewest layers :math:`\\sigma_{1}, \\, \\sigma_{2}, \\, \\cdots, \\, \\sigma_{t}` such that:
        - each layer has at most :math:`m` disjoint transpositions
        - each layer moves a power‑of‑two number of indices.

    We do this as follows:
        1) Cycle extraction – split sigma into disjoint cycles.
        2) Star factorisation – a k‑cycle becomes (k-1) hub–spoke swaps.
        3) Greedy packing – merge swaps into layers while keeping rules.

    Args:
        sigma (list[int] or tuple[int]): permutation description on :math:`\\{0, \\, 1, \\, \\dots, \\, d-1\\}`.
        m (int): power‑of‑two budget.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be used
            in the execution. If ``None``, it uses the current backend. Defaults to ``None``.

    Returns:
        list[list[tuple[int, int]]]: :math:`t` layers of pairwise transpositions.

    """
    backend = _check_backend(backend)

    if isinstance(sigma, tuple):
        sigma = list(sigma)

    if not isinstance(sigma, (list, tuple)):
        raise_error(
            TypeError, f"Permutation sigma must be ``list`` or ``tuple`` of ``int``s."
        )

    if sum([abs(s - i) for s, i in zip(sorted(sigma), range(len(sigma)))]) != 0:
        raise_error(
            ValueError, "Permutation sigma must contain all indices {0,...,n-1}"
        )

    if m > 0 and (m & (m - 1)) != 0:
        raise_error(ValueError, f"budget m must be a power‑of‑two")

    matchings = [l for cyc in _cycles_from_perm(sigma) for l in _star_matchings(cyc)]

    return _greedy_pack(matchings, m)



from typing import Optional

import networkx as nx

from qibo import gates
from qibo.backends import _check_backend_and_local_state
from qibo.config import raise_error
from qibo.models import Circuit
from qibo.transpiler._exceptions import ConnectivityError, PlacementError
from qibo.transpiler.abstract import Placer, Router
from qibo.transpiler.asserts import assert_placement
from qibo.transpiler.router import _find_connected_qubit


def _find_gates_qubits_pairs(circuit: Circuit):
    """Helper method for :meth:`qibo.transpiler.placer`.
    Translate circuit into a list of pairs of qubits to be used by the router and placer.

    Args:
        circuit (:class:`qibo.models.circuit.Circuit`): Circuit to be translated.

    Returns:
        (list): Pairs of qubits targeted by two qubits gates.
    """
    gates_qubits_pairs = []
    for gate in circuit.queue:
        if isinstance(gate, gates.M):
            pass
        elif len(gate.qubits) == 2:
            gates_qubits_pairs.append(sorted(gate.qubits))
        elif len(gate.qubits) >= 3:
            raise_error(
                ValueError, "Gates targeting more than 2 qubits are not supported"
            )
    return gates_qubits_pairs


class StarConnectivityPlacer(Placer):
    """Find an optimized qubit placement for the following connectivity:

             q
             |
        q -- q -- q
             |
             q

    Args:
        connectivity (:class:`networkx.Graph`): Star connectivity graph.
    """

    def __init__(self, connectivity: Optional[nx.Graph] = None):
        self.connectivity = connectivity
        self.middle_qubit = None

    def __call__(self, circuit: Circuit):
        """Apply the transpiler transformation on a given circuit.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): The original Qibo circuit to transform.
                Only single qubit gates and two qubits gates are supported by the router.
        """
        assert_placement(circuit, self.connectivity)
        self._check_star_connectivity()

        middle_qubit_idx = circuit.wire_names.index(self.middle_qubit)
        wire_names = circuit.wire_names.copy()

        for i, gate in enumerate(circuit.queue):
            if len(gate.qubits) > 2:
                raise_error(
                    PlacementError,
                    "Gates targeting more than 2 qubits are not supported",
                )
            if len(gate.qubits) == 2:
                if middle_qubit_idx not in gate.qubits:
                    new_middle = _find_connected_qubit(
                        gate.qubits,
                        circuit.queue[i + 1 :],
                        error=PlacementError,
                        mapping=list(range(circuit.nqubits)),
                    )

                    (
                        wire_names[middle_qubit_idx],
                        wire_names[new_middle],
                    ) = (
                        wire_names[new_middle],
                        wire_names[middle_qubit_idx],
                    )
                    break

        circuit.wire_names = wire_names

    def _check_star_connectivity(self):
        """Check if the connectivity graph is a star graph."""
        if len(self.connectivity.nodes) != 5:
            raise_error(
                ConnectivityError,
                f"This connectivity graph is not a star graph. Length of nodes provided: {len(self.connectivity.nodes)} != 5.",
            )
        for node in self.connectivity.nodes:
            if self.connectivity.degree(node) == 4:
                self.middle_qubit = node
            elif self.connectivity.degree(node) != 1:
                raise_error(
                    ConnectivityError,
                    "This connectivity graph is not a star graph. There is a node with degree different from 1 and 4.",
                )


class Subgraph(Placer):
    """
    Subgraph isomorphism qubit placer.

    Since it is a :math:`NP`-complete problem, it can take exponential time for large circuits.
    This initialization method may fail for very short circuits.

    Attributes:
        connectivity (:class:`networkx.Graph`): Hardware connectivity.
    """

    def __init__(self, connectivity: Optional[nx.Graph] = None):
        self.connectivity = connectivity

    def __call__(self, circuit: Circuit):
        """Find the initial layout of the given circuit using subgraph isomorphism.
        Circuit must contain at least two two-qubit gates to implement subgraph placement.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): Circuit to be transpiled.
        """
        assert_placement(circuit, self.connectivity)
        gates_qubits_pairs = _find_gates_qubits_pairs(circuit)
        if len(gates_qubits_pairs) < 2:
            raise_error(
                ValueError,
                "Circuit must contain at least two two-qubit gates "
                + "to implement subgraph placement.",
            )
        circuit_subgraph = nx.Graph()
        circuit_subgraph.add_nodes_from(list(range(circuit.nqubits)))
        matcher = nx.algorithms.isomorphism.GraphMatcher(
            self.connectivity, circuit_subgraph
        )
        i = 0
        circuit_subgraph.add_edge(gates_qubits_pairs[i][0], gates_qubits_pairs[i][1])
        while matcher.subgraph_is_monomorphic():
            result = matcher
            i += 1
            circuit_subgraph.add_edge(
                gates_qubits_pairs[i][0], gates_qubits_pairs[i][1]
            )
            matcher = nx.algorithms.isomorphism.GraphMatcher(
                self.connectivity, circuit_subgraph
            )
            if (
                self.connectivity.number_of_edges()
                == circuit_subgraph.number_of_edges()
                or i == len(gates_qubits_pairs) - 1
            ):
                break

        circuit.wire_names = sorted(result.mapping, key=lambda k: result.mapping[k])


class Random(Placer):
    """
    Random initialization with greedy policy, let a maximum number of 2-qubit
    gates can be applied without introducing any SWAP gate.

    Attributes:
        connectivity (:class:`networkx.Graph`): Hardware connectivity.
        samples (int, optional): Number of random initializations to try.
            Defaults to :math:`100`.
        seed (int or :class:`numpy.random.Generator`, optional): Either a generator of
            random numbers or a fixed seed to initialize a generator. If ``None``,
            initializes a generator with a random seed. Defaults to ``None``.
    """

    def __init__(
        self, connectivity: Optional[nx.Graph] = None, samples: int = 100, seed=None
    ):
        self.connectivity = connectivity
        self.samples = samples
        self.seed = seed

    def __call__(self, circuit):
        """Find an initial layout of the given circuit using random greedy algorithm.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): Circuit to be transpiled.
        """
        assert_placement(circuit, self.connectivity)
        _, local_state = _check_backend_and_local_state(self.seed, backend=None)
        gates_qubits_pairs = _find_gates_qubits_pairs(circuit)
        nodes = self.connectivity.number_of_nodes()
        keys = list(self.connectivity.nodes())

        final_mapping = dict(zip(keys, range(nodes)))
        final_graph = nx.relabel_nodes(self.connectivity, final_mapping)
        final_cost = self._cost(final_graph, gates_qubits_pairs)
        for _ in range(self.samples):
            mapping = dict(
                zip(keys, local_state.choice(range(nodes), nodes, replace=False))
            )
            graph = nx.relabel_nodes(self.connectivity, mapping)
            cost = self._cost(graph, gates_qubits_pairs)

            if cost == 0:
                final_layout = dict(zip(keys, list(mapping.values())))
                circuit.wire_names = sorted(final_layout, key=final_layout.get)
                return

            if cost < final_cost:
                final_graph = graph
                final_mapping = mapping
                final_cost = cost

        final_layout = dict(zip(keys, list(final_mapping.values())))
        circuit.wire_names = sorted(final_layout, key=final_layout.get)

    def _cost(self, graph: nx.Graph, gates_qubits_pairs: list):
        """
        Compute the cost associated to an initial layout as the lengh of the reduced circuit.

        Args:
            graph (:class:`networkx.Graph`): Hardware connectivity.
            gates_qubits_pairs (list): Circuit representation.

        Returns:
            (int): lengh of the reduced circuit.
        """
        for allowed, gate in enumerate(gates_qubits_pairs):
            if gate not in graph.edges():
                return len(gates_qubits_pairs) - allowed - 1

        return 0


class ReverseTraversal(Placer):
    """
    Places qubits based on the algorithm proposed in Reference [1].

    Compatible with all the available ``Router``s.

    Args:
        connectivity (:class:`networkx.Graph`): Hardware connectivity.
        routing_algorithm (:class:`qibo.transpiler.abstract.Router`): Router to be used.
        depth (int, optional): Number of two-qubit gates to be considered for routing.
            If ``None`` just one backward step will be implemented.
            If depth is greater than the number of two-qubit gates in the circuit,
            the circuit will be routed more than once.
            Example: on a circuit with four two-qubit gates :math:`A-B-C-D`
            using depth :math:`d = 6`, the routing will be performed
            on the circuit :math:`C-D-D-C-B-A`.

    References:
        1. G. Li, Y. Ding, and Y. Xie,
        *Tackling the Qubit Mapping Problem for NISQ-Era Quantum Devices*.
        `arXiv:1809.02573 [cs.ET] <https://arxiv.org/abs/1809.02573>`_.
    """

    def __init__(
        self,
        routing_algorithm: Router,
        connectivity: Optional[nx.Graph] = None,
        depth: Optional[int] = None,
    ):
        self.connectivity = connectivity
        self.routing_algorithm = routing_algorithm
        self.depth = depth

    def __call__(self, circuit: Circuit):
        """Find the initial layout of the given circuit using Reverse Traversal placement.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): Circuit to be transpiled.
        """
        assert_placement(circuit, self.connectivity)
        self.routing_algorithm.connectivity = self.connectivity
        new_circuit = self._assemble_circuit(circuit)
        self._routing_step(new_circuit)

    def _assemble_circuit(self, circuit: Circuit):
        """Assemble a single circuit to apply Reverse Traversal placement based on depth.

        Example: for a circuit with four two-qubit gates :math:`A-B-C-D`
        using depth :math:`d = 6`, the function will return the circuit :math:`C-D-D-C-B-A`.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): Circuit to be assembled.

        Returns:
            (:class:`qibo.models.circuit.Circuit`): Assembled circuit to perform Reverse Traversal placement.
        """

        if self.depth is None:
            return circuit.invert()

        gates_qubits_pairs = _find_gates_qubits_pairs(circuit)
        circuit_gates = len(gates_qubits_pairs)
        if circuit_gates == 0:
            raise_error(
                ValueError, "The circuit must contain at least a two-qubit gate."
            )
        repetitions, remainder = divmod(self.depth, circuit_gates)

        assembled_gates_qubits_pairs = []
        for _ in range(repetitions):
            assembled_gates_qubits_pairs += gates_qubits_pairs[:]
            gates_qubits_pairs.reverse()
        assembled_gates_qubits_pairs += gates_qubits_pairs[0:remainder]

        new_circuit = Circuit(circuit.nqubits, wire_names=circuit.wire_names)
        for qubits in assembled_gates_qubits_pairs:
            # As only the connectivity is important here we can replace everything with CZ gates
            new_circuit.add(gates.CZ(qubits[0], qubits[1]))

        return new_circuit.invert()

    def _routing_step(self, circuit: Circuit):
        """Perform routing of the circuit.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): Circuit to be routed.
        """
        _, final_mapping = self.routing_algorithm(circuit)

        return final_mapping



from functools import cache
from inspect import signature
from itertools import product
from typing import List, Union

import numpy as np
from sympy import S

from qibo import Circuit, gates, symbols
from qibo.backends import _check_backend
from qibo.config import raise_error
from qibo.hamiltonians import SymbolicHamiltonian
from qibo.transpiler.optimizer import Preprocessing
from qibo.transpiler.pipeline import Passes
from qibo.transpiler.placer import Random
from qibo.transpiler.router import Sabre
from qibo.transpiler.unroller import NativeGates, Unroller

SUPPORTED_NQUBITS = [1, 2]
"""Supported nqubits for GST."""


def _check_nqubits(nqubits):
    if nqubits not in SUPPORTED_NQUBITS:
        raise_error(
            ValueError,
            f"nqubits given as {nqubits}. nqubits needs to be either 1 or 2.",
        )


@cache
def _gates(nqubits) -> List:
    """Gates implementing all the GST state preparations.

    Args:
        nqubits (int): Number of qubits for the circuit.
    Returns:
        List(:class:`qibo.gates.Gate`): gates used to prepare the possible states.
    """

    return list(
        product(
            [(gates.I,), (gates.X,), (gates.H,), (gates.H, gates.S)], repeat=nqubits
        )
    )


@cache
def _measurements(nqubits: int) -> List:
    """Measurement gates implementing all the GST measurement bases.

    Args:
        nqubits (int): Number of qubits for the circuit.
    Returns:
        List(:class:`qibo.gates.Gate`): gates implementing the possible measurement bases.
    """

    return list(product([gates.Z, gates.X, gates.Y, gates.Z], repeat=nqubits))


@cache
def _observables(nqubits: int) -> List:
    """All the observables measured in the GST protocol.

    Args:
        nqubits (int): number of qubits for the circuit.

    Returns:
        List[:class:`qibo.symbols.Symbol`]: all possible observables to be measured.
    """

    return list(product([symbols.I, symbols.Z, symbols.Z, symbols.Z], repeat=nqubits))


@cache
def _get_observable(j: int, nqubits: int):
    """Returns the :math:`j`-th observable. The :math:`j`-th observable is expressed as a base-4 indexing and is given by

    .. math::
        j \\in \\{0, 1, 2, 3\\}^{\\otimes n} \\equiv \\{ I, X, Y, Z\\}^{\\otimes n}.

    Args:
        j (int): index of the measurement basis (in base-4)
        nqubits (int): number of qubits.

    Returns:
        List[:class:`qibo.hamiltonians.SymbolicHamiltonian`]: observables represented by symbolic Hamiltonians.
    """

    if j == 0:
        _check_nqubits(nqubits)
    observables = _observables(nqubits)[j]
    observable = S(1)
    for q, obs in enumerate(observables):
        if obs is not symbols.I:
            observable *= obs(q)
    return SymbolicHamiltonian(observable, nqubits=nqubits)


@cache
def _prepare_state(k, nqubits):
    """Prepares the :math:`k`-th state for an :math:`n`-qubits (`nqubits`) circuit.
    Using base-4 indexing for :math:`k`,

    .. math::
        k \\in \\{0, 1, 2, 3\\}^{\\otimes n} \\equiv \\{ 0\\rangle\\langle0|, |1\\rangle\\langle1|,
        |+\\rangle\\langle +|, |y+\\rangle\\langle y+|\\}^{\\otimes n}.

    Args:
        k (int): index of the state to be prepared.
        nqubits (int): Number of qubits.

    Returns:
        list(:class:`qibo.gates.Gate`): gates that prepare the :math:`k`-th state.
    """

    _check_nqubits(nqubits)
    gates = _gates(nqubits)[k]
    return [gate(q) for q in range(len(gates)) for gate in gates[q]]


@cache
def _measurement_basis(j, nqubits):
    """Constructs the :math:`j`-th measurement basis element for an :math:`n`-qubits (`nqubits`) circuit.
    Base-4 indexing is used for the :math:`j`-th measurement basis and is given by

    .. math::
        j \\in \\{0, 1, 2, 3\\}^{\\otimes n} \\equiv \\{ I, X, Y, Z\\}^{\\otimes n}.

    Args:
        j (int): index of the measurement basis element.
        nqubits (int): number of qubits.

    Returns:
        List[:class:`qibo.gates.Gate`]: gates forming the :math:`j`-th element
            of the Pauli measurement basis.
    """

    _check_nqubits(nqubits)
    measurements = _measurements(nqubits)[j]
    return [gates.M(q, basis=measurements[q]) for q in range(len(measurements))]


def _gate_tomography(
    nqubits: int,
    gate: gates.Gate = None,
    nshots: int = int(1e4),
    noise_model=None,
    backend=None,
    transpiler=None,
):
    """Runs gate tomography for a 1 or 2 qubit gate.

    It obtains a :math:`4^{n} \\times 4^{n}` matrix, where :math:`n` is the number of qubits.
    This matrix needs to be post-processed to get the Pauli-Liouville representation of the gate.
    The matrix has elements :math:`\\text{tr}(M_{j} \\, \\rho_{k})` or
    :math:`\\text{tr}(M_{j} \\, O_{l} \\rho_{k})`, depending on whether the gate
    :math:`O_{l}` is present or not.

    Args:
        nqubits (int): number of qubits of the gate.
        gate (:class:`qibo.gates.Gate`, optional): gate to perform gate tomography on.
            If ``None``, then gate tomography will be performed for an empty circuit.
            Defaults to ``None``.
        nshots (int, optional): number of shots used.
        noise_model (:class:`qibo.noise.NoiseModel`, optional): noise model applied to simulate
            noisy computations.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend
            to be used in the execution. If ``None``, it uses
            the current backend. Defaults to ``None``.

    Returns:
        ndarray: matrix approximating the input gate.
    """

    # Check if gate is 1 or 2 qubit gate.
    _check_nqubits(nqubits)

    backend = _check_backend(backend)

    if gate is not None:
        if nqubits != len(gate.qubits):
            raise_error(
                ValueError,
                f"Mismatched inputs: nqubits given as {nqubits}. {gate} is a {len(gate.qubits)}-qubit gate.",
            )
        gate = gate.__class__(*gate.qubits, **gate.init_kwargs)

    # GST for empty circuit or with gates
    matrix_jk = 1j * np.zeros((4**nqubits, 4**nqubits))
    for k in range(4**nqubits):
        circ = Circuit(nqubits, density_matrix=True)
        circ.add(_prepare_state(k, nqubits))

        if gate is not None:
            circ.add(gate)

        for j in range(4**nqubits):
            if j == 0:
                exp_val = 1.0
            else:
                new_circ = circ.copy()
                measurements = _measurement_basis(j, nqubits)
                new_circ.add(measurements)
                observable = _get_observable(j, nqubits)
                if noise_model is not None and backend.name != "qibolab":
                    new_circ = noise_model.apply(new_circ)
                if transpiler is not None:
                    new_circ, _ = transpiler(new_circ)
                exp_val = observable.expectation_from_samples(
                    backend.execute_circuit(new_circ, nshots=nshots).frequencies()
                )
            matrix_jk[j, k] = exp_val
    return backend.cast(matrix_jk, dtype=matrix_jk.dtype)


def GST(
    gate_set: Union[tuple, set, list],
    nshots=int(1e4),
    noise_model=None,
    include_empty=False,
    pauli_liouville=False,
    gauge_matrix=None,
    backend=None,
    transpiler=None,
):
    """Run Gate Set Tomography on the input ``gate_set``.

    Args:
        gate_set (tuple or set or list): set of :class:`qibo.gates.Gate` and parameters to run
            GST on. For instance, ``gate_set = [(gates.RX, [np.pi/3]), gates.Z, (gates.PRX,
            [np.pi/2, np.pi/3]), (gates.GPI, [np.pi/7]), gates.CNOT]``.
        nshots (int, optional): number of shots used in Gate Set Tomography per gate.
            Defaults to :math:`10^{4}`.
        noise_model (:class:`qibo.noise.NoiseModel`, optional): noise model applied to simulate
            noisy computations.
        include_empty (bool, optional): if ``True``, additionally performs gate set tomography
            for :math:`1`- and :math:`2`-qubit empty circuits, returning the corresponding empty
            matrices in the first and second position of the ouput list.
        pauli_liouville (bool, optional): if ``True``, returns the matrices in the
            Pauli-Liouville representation. Defaults to ``False``.
        gauge_matrix (ndarray, optional): gauge matrix transformation to the Pauli-Liouville
            representation. Defaults to

            .. math::
                \\begin{pmatrix}
                    1 & 1 & 1 & 1 \\\\
                    0 & 0 & 1 & 0 \\\\
                    0 & 0 & 0 & 1 \\\\
                    1 & -1 & 0 & 0 \\\\
                \\end{pmatrix}

        backend (:class:`qibo.backends.abstract.Backend`, optional): backend
            to be used in the execution. If ``None``, it uses
            the current backend. Defaults to ``None``.


    Returns:
        List(ndarray): input ``gate_set`` represented by matrices estimaded via GST.
    """

    backend = _check_backend(backend)

    if backend.name == "qibolab" and transpiler is None:  # pragma: no cover
        transpiler = Passes(
            connectivity=backend.platform.topology,
            passes=[
                Preprocessing(backend.platform.topology),
                Random(backend.platform.topology),
                Sabre(backend.platform.topology),
                Unroller(NativeGates.default()),
            ],
        )

    matrices = []
    empty_matrices = []
    if include_empty or pauli_liouville:
        for nqubits in SUPPORTED_NQUBITS:
            empty_matrix = _gate_tomography(
                nqubits=nqubits,
                gate=None,
                nshots=nshots,
                noise_model=noise_model,
                backend=backend,
                transpiler=transpiler,
            )
            empty_matrices.append(empty_matrix)

    for gate in gate_set:
        if gate is not None:

            if isinstance(gate, tuple):
                angles = ["theta", "phi", "lam"]
                gate, params = gate
                init_args = signature(gate).parameters
                valid_angles = [arg for arg in init_args if arg in angles]
                angle_values = dict(zip(valid_angles, params))
            else:
                angle_values = {}
                init_args = signature(gate).parameters

            if "q" in init_args:
                nqubits = 1
            elif "q0" in init_args and "q1" in init_args and "q2" not in init_args:
                nqubits = 2
            else:
                raise_error(
                    RuntimeError,
                    f"Gate {gate} is not supported for `GST`, only 1- and 2-qubit gates are supported.",
                )
            gate = gate(*range(nqubits), **angle_values)

        matrices.append(
            _gate_tomography(
                nqubits=nqubits,
                gate=gate,
                nshots=nshots,
                noise_model=noise_model,
                backend=backend,
                transpiler=transpiler,
            )
        )

    if pauli_liouville:
        if gauge_matrix is not None:
            if np.linalg.det(gauge_matrix) == 0:
                raise_error(ValueError, "Matrix is not invertible")
        else:
            gauge_matrix = backend.cast(
                [[1, 1, 1, 1], [0, 0, 1, 0], [0, 0, 0, 1], [1, -1, 0, 0]]
            )
        PL_matrices = []
        gauge_matrix_1q = gauge_matrix
        gauge_matrix_2q = backend.np.kron(gauge_matrix, gauge_matrix)
        for matrix in matrices:
            gauge_matrix = gauge_matrix_1q if matrix.shape[0] == 4 else gauge_matrix_2q
            empty = empty_matrices[0] if matrix.shape[0] == 4 else empty_matrices[1]
            PL_matrices.append(
                gauge_matrix
                @ backend.np.linalg.inv(empty)
                @ matrix
                @ backend.np.linalg.inv(gauge_matrix)
            )
        matrices = PL_matrices

    if include_empty:
        matrices = empty_matrices + matrices

    return matrices



from typing import Optional

from qibo.backends import Backend, _check_backend
from qibo.config import raise_error
from qibo.quantum_info.utils import _get_single_paulis, _pauli_basis_normalization


def pauli_basis(
    nqubits: int,
    normalize: bool = False,
    vectorize: bool = False,
    sparse: bool = False,
    order: Optional[str] = None,
    pauli_order: str = "IXYZ",
    backend: Optional[Backend] = None,
):
    """Creates the ``nqubits``-qubit Pauli basis.

    Args:
        nqubits (int): number of qubits.
        normalize (bool, optional): If ``True``, normalized basis is returned.
            Defaults to False.
        vectorize (bool, optional): If ``False``, returns a nested array with
            all Pauli matrices. If ``True``, retuns an array where every
            row is a vectorized Pauli matrix. Defaults to ``False``.
        sparse (bool, optional): If ``True``, retuns Pauli basis in a sparse
            representation. Defaults to ``False``.
        order (str, optional): If ``"row"``, vectorization of Pauli basis is
            performed row-wise. If ``"column"``, vectorization is performed
            column-wise. If ``"system"``, system-wise vectorization is
            performed. If ``vectorization=False``, then ``order=None`` is
            forced. Defaults to ``None``.
        pauli_order (str, optional): corresponds to the order of 4 single-qubit
            Pauli elements. Defaults to ``"IXYZ"``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend
            to be used in the execution. If ``None``, it uses
            the current backend. Defaults to ``None``.

    Returns:
        ndarray or tuple: all Pauli matrices forming the basis. If ``sparse=True``
            and ``vectorize=True``, tuple is composed of an array of non-zero
            elements and an array with their row-wise indexes.
    """

    if set(pauli_order) != {"I", "X", "Y", "Z"}:
        raise_error(
            ValueError,
            f"pauli_order has to contain 4 symbols: I, X, Y, Z. Got {pauli_order} instead.",
        )

    if vectorize and order is None:
        raise_error(ValueError, "when vectorize=True, order must be specified.")

    if sparse and not vectorize:
        raise_error(
            NotImplementedError,
            "sparse representation is not implemented for unvectorized Pauli basis.",
        )

    backend = _check_backend(backend)
    fname = f"_pauli_basis_{order}"
    normalization = _pauli_basis_normalization(nqubits) if normalize else 1.0

    if vectorize and sparse:
        func = getattr(backend.qinfo, f"_vectorize_sparse{fname}")
    elif vectorize:
        func = getattr(backend.qinfo, f"_vectorize{fname}")
    else:
        func = backend.qinfo._pauli_basis

    return func(
        nqubits, *_get_single_paulis(pauli_order, backend), normalization=normalization
    )


def comp_basis_to_pauli(
    nqubits: int,
    normalize: bool = False,
    sparse: bool = False,
    order: str = "row",
    pauli_order: str = "IXYZ",
    backend=None,
):
    """Unitary matrix :math:`U` that converts operators from the Liouville
    representation in the computational basis to the Pauli-Liouville
    representation.

    The unitary :math:`U` is given by

    .. math::
        U = \\sum_{k = 0}^{d^{2} - 1} \\, |k)(P_{k}| \\,\\, ,

    where :math:`|P_{k})` is the vectorization of the :math:`k`-th
    Pauli operator :math:`P_{k}`, and :math:`|k)` is the vectorization
    of the :math:`k`-th computational basis element.
    For a definition of vectorization, see :func:`qibo.quantum_info.vectorization`.

    Example:
        .. code-block:: python

            from qibo.quantum_info import random_density_matrix, vectorization, comp_basis_to_pauli
            nqubits = 2
            d = 2**nqubits
            rho = random_density_matrix(d)
            U_c2p = comp_basis_to_pauli(nqubits)
            rho_liouville = vectorization(rho, order="system")
            rho_pauli_liouville = U_c2p @ rho_liouville

    Args:
        nqubits (int): number of qubits.
        normalize (bool, optional): If ``True``, converts to the
            Pauli basis. Defaults to ``False``.
        sparse (bool, optional): If ``True``, returns unitary matrix in
            sparse representation. Defaults to ``False``.
        order (str, optional): If ``"row"``, vectorization of Pauli basis is
            performed row-wise. If ``"column"``, vectorization is performed
            column-wise. If ``"system"``, system-wise vectorization is
            performed. Defaults to ``"row"``.
        pauli_order (str, optional): corresponds to the order of 4 single-qubit
            Pauli elements. Defaults to ``"IXYZ"``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
            used in the execution. If ``None``, it uses
            the current backend. Defaults to ``None``.

    Returns:
        ndarray or tuple: Unitary matrix :math:`U`. If ``sparse=True``,
            tuple is composed of array of non-zero elements and an
            array with their row-wise indexes.

    """
    backend = _check_backend(backend)

    if sparse:
        elements, indexes = pauli_basis(
            nqubits,
            normalize,
            vectorize=True,
            sparse=sparse,
            order=order,
            pauli_order=pauli_order,
            backend=backend,
        )
        elements = backend.np.conj(elements)

        return elements, indexes

    unitary = pauli_basis(
        nqubits,
        normalize,
        vectorize=True,
        sparse=sparse,
        order=order,
        pauli_order=pauli_order,
        backend=backend,
    )

    unitary = backend.np.conj(unitary)

    return unitary


def pauli_to_comp_basis(
    nqubits: int,
    normalize: bool = False,
    sparse: bool = False,
    order: str = "row",
    pauli_order: str = "IXYZ",
    backend=None,
):
    """Unitary matrix :math:`U` that converts operators from the
    Pauli-Liouville representation to the Liouville representation
    in the computational basis.

    The unitary :math:`U` is given by

    .. math::
        U = \\sum_{k = 0}^{d^{2} - 1} \\, |P_{k})(b_{k}| \\, ,

    where :math:`|P_{k})` is the vectorization of the :math:`k`-th
    Pauli operator :math:`P_{k}`, and :math:`|k)` is the vectorization
    of the :math:`k`-th computational basis element.
    For a definition of vectorization, see :func:`qibo.quantum_info.vectorization`.

    Args:
        nqubits (int): number of qubits.
        normalize (bool, optional): If ``True``, converts to the
            Pauli basis. Defaults to ``False``.
        sparse (bool, optional): If ``True``, returns unitary matrix in
            sparse representation. Defaults to ``False``.
        order (str, optional): If ``"row"``, vectorization of Pauli basis is
            performed row-wise. If ``"column"``, vectorization is performed
            column-wise. If ``"system"``, system-wise vectorization is
            performed. Defaults to ``"row"``.
        pauli_order (str, optional): corresponds to the order of 4 single-qubit
            Pauli elements. Defaults to ``"IXYZ"``.
        backend (:class:`qibo.backends.abstract.Backend`, optional): backend to be
            used in the execution. If ``None``, it uses
            the current backend. Defaults to ``None``.

    Returns:
        ndarray or tuple: Unitary matrix :math:`U`. If ``sparse=True``,
            tuple is composed of array of non-zero elements and an
            array with their row-wise indexes.
    """
    backend = _check_backend(backend)

    if sparse:
        normalization = _pauli_basis_normalization(nqubits) if normalize else 1.0
        func = getattr(backend.qinfo, f"_pauli_to_comp_basis_sparse_{order}")
        return func(
            nqubits,
            *_get_single_paulis(pauli_order, backend),
            normalization=normalization,
        )

    return pauli_basis(
        nqubits,
        normalize,
        vectorize=True,
        sparse=False,
        order=order,
        pauli_order=pauli_order,
        backend=backend,
    ).T



import networkx as nx

from qibo.config import raise_error
from qibo.models import Circuit
from qibo.transpiler._exceptions import (
    ConnectivityError,
    PlacementError,
    TranspilerPipelineError,
)
from qibo.transpiler.abstract import Optimizer, Placer, Router
from qibo.transpiler.asserts import (
    assert_connectivity,
    assert_decomposition,
    assert_placement,
)
from qibo.transpiler.unroller import DecompositionError, NativeGates, Unroller


def restrict_connectivity_qubits(connectivity: nx.Graph, qubits: list[str]):
    """Restrict the connectivity to selected qubits.

    Args:
        connectivity (:class:`networkx.Graph`): Hardware connectivity.
        qubits (list): List of qubits to select.

    Returns:
        (:class:`networkx.Graph`): New restricted connectivity.
    """
    if not set(qubits).issubset(set(connectivity.nodes)):
        raise_error(
            ConnectivityError, "Some qubits are not in the original connectivity."
        )

    new_connectivity = nx.Graph()
    new_connectivity.add_nodes_from(qubits)
    new_edges = [
        edge for edge in connectivity.edges if edge[0] in qubits and edge[1] in qubits
    ]
    new_connectivity.add_edges_from(new_edges)

    if not nx.is_connected(new_connectivity):
        raise_error(ConnectivityError, "New connectivity graph is not connected.")

    return new_connectivity


class Passes:
    """Define a transpiler pipeline consisting of smaller transpiler steps that are applied sequentially:

    Args:
        passes (list, optional): List of transpiler passes to be applied sequentially.
            If ``None``, default transpiler will be used. Defaults to ``None``.
        connectivity (:class:`networkx.Graph`, optional): Hardware connectivity.
        native_gates (:class:`qibo.transpiler.unroller.NativeGates`, optional): Native gates supported by the hardware.
            Defaults to :class:`qibo.transpiler.unroller.NativeGates.default()`.
        on_qubits (list, optional): List of qubits to be used in the transpiler.
            If ``None``, all qubits in the connectivity will be used. Defaults to ``None``.
    """

    def __init__(
        self,
        passes: list = None,
        connectivity: nx.Graph = None,
        native_gates: NativeGates = NativeGates.default(),
        on_qubits: list = None,
    ):
        if on_qubits is not None:
            connectivity = restrict_connectivity_qubits(connectivity, on_qubits)
        self.connectivity = connectivity
        self.native_gates = native_gates
        self.passes = [] if passes is None else passes

    def __call__(self, circuit):
        """Apply the transpiler pipeline to the circuit.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): Circuit to be transpiled.

        Returns:
            (:class:`qibo.models.circuit.Circuit`, dict): Transpiled circuit and final {logical: physical} qubit mapping.
        """

        final_layout = None
        for transpiler_pass in self.passes:
            if isinstance(transpiler_pass, Optimizer):
                transpiler_pass.connectivity = self.connectivity
                circuit = transpiler_pass(circuit)
            elif isinstance(transpiler_pass, Placer):
                transpiler_pass.connectivity = self.connectivity
                final_layout = transpiler_pass(circuit)
            elif isinstance(transpiler_pass, Router):
                transpiler_pass.connectivity = self.connectivity
                circuit, final_layout = transpiler_pass(circuit)
            elif isinstance(transpiler_pass, Unroller):
                circuit = transpiler_pass(circuit)
            else:
                raise_error(
                    TranspilerPipelineError,
                    f"Unrecognised transpiler pass: {transpiler_pass}",
                )
        return circuit, final_layout

    def is_satisfied(self, circuit: Circuit):
        """Check if the circuit respects the hardware connectivity and native gates.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): Circuit to be checked.

        Returns:
            (bool): ``True`` if the circuit respects the hardware connectivity and native gates, ``False`` otherwise.
        """
        try:
            assert_placement(circuit=circuit, connectivity=self.connectivity)
            assert_connectivity(circuit=circuit, connectivity=self.connectivity)
            assert_decomposition(circuit=circuit, native_gates=self.native_gates)
            return True
        except (ConnectivityError, DecompositionError, PlacementError, ValueError):
            return False



from qibo.gates.abstract import Gate

QIBO_DEFAULT_COLOR = "#C194D8"
QIBO_COMPLEMENTARY_COLOR = "#DE8004"


class FusedStartGateBarrier(Gate):
    """
    :class:`qibo.ui.drawer_utils.FusedStartGateBarrier` gives room to fused group of gates.
    Inherit from ``qibo.gates.abstract.Gate``. A special gate barrier gate to pin the starting point of fused gates.
    """

    def __init__(self, q_ctrl, q_trgt, nfused, equal_qbits=False):

        super().__init__()
        self.name = (
            "FusedStartGateBarrier"
            + str(nfused)
            + ("" if not equal_qbits else "@EQUAL")
        )
        self.draw_label = ""
        self.control_qubits = (q_ctrl,)
        self.target_qubits = (q_trgt,) if q_ctrl != q_trgt else ()
        self.init_args = [q_trgt, q_ctrl] if q_ctrl != q_trgt else [q_ctrl]
        self.unitary = False
        self.is_controlled_by = False
        self.nfused = nfused


class FusedEndGateBarrier(Gate):
    """
    :class:`qibo.ui.drawer_utils.FusedEndGateBarrier` gives room to fused group of gates.
    Inherit from ``qibo.gates.abstract.Gate``. A special gate barrier gate to pin the ending point of fused gates.
    """

    def __init__(self, q_ctrl, q_trgt):

        super().__init__()
        self.name = "FusedEndGateBarrier"
        self.draw_label = ""
        self.control_qubits = (q_ctrl,)
        self.target_qubits = (q_trgt,) if q_ctrl != q_trgt else ()
        self.init_args = [q_trgt, q_ctrl] if q_ctrl != q_trgt else [q_ctrl]
        self.unitary = False
        self.is_controlled_by = False


def generate_bitstring_combinations(n):
    """Generate all bitstring combinations given bitstring length `n`."""
    bitstrings = []
    for i in range(2**n):
        bitstrings.append(format(i, f"0{n}b"))
    return bitstrings
